// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package app

import (
	"context"
	"fmt"
	"os"
	"os/signal"
	"runtime"
	"runtime/debug"
	"strings"
	"sync/atomic"
	"syscall"
	"time"

	"hello-api/internal/hinting"

	"hello-api/internal/lct"
	"hello-api/internal/logger"
	"hello-api/internal/metrics"
	"hello-api/internal/utils"
	"hello-api/internal/utils/streamchecker"

	"hello-api/internal/config"

	"github.com/aws/aws-sdk-go-v2/service/kinesis"
	producer "github.com/zacharyestep/kinesis-producer"
	"go.uber.org/zap"
)

// HelloService -- Contains all the components required to execute the service
type HelloService struct {
	cfg           *config.Config
	HttpServer    *HttpServer
	metricsClient metrics.Service
	alive         *int32
	Ready         *int32
	signals       <-chan os.Signal
	Logger        logger.Logger
	kinesisClient *kinesis.Client
	lctProducer   lct.StartStopableLctProducer
	hintsClient   hinting.Hinter
}

// NewHelloService -- Create a new hello service from a default configuration.
func NewHelloService() *HelloService {
	hs := NewHelloServiceFromCfg(config.GetInstance(), hookSignals())
	logger.BaseLogger().Info(fmt.Sprintf("Configuration loaded:\n  %+v\n--------------------------------------------------", hs.cfg))
	return hs
}

// NewHelloServiceFromCfg -- Create a new hello service from a supplied configuration.
func NewHelloServiceFromCfg(cfg *config.Config, signals <-chan os.Signal) *HelloService {
	return &HelloService{Logger: logger.BaseLogger(), cfg: cfg, Ready: new(int32), alive: new(int32), signals: signals}
}

func (service *HelloService) setupAndRun() (err error) {
	metricsClient, wfErr := metrics.NewClient(service.cfg)
	if wfErr != nil {
		logger.BaseLogger().Warn("Unable to create WF client, metrics support will be disabled for Wavefront", zap.Error(wfErr))
	}

	logLevel := config.GetLoggerConfigInstance().LogLevel
	if logLevel == "debug" && metricsClient != nil {
		metrics.RegisterGoStatMetrics(metricsClient)
	}
	service.metricsClient = metricsClient

	// Initialize our AWS Kinesis client
	service.Logger.Debug("Creating kinesis client",
		zap.String("AwsEndpoint", service.cfg.AwsEndpoint), zap.String("Region", service.cfg.Region))
	service.kinesisClient, err = config.CreateKinesisClient(service.cfg)
	if err != nil {
		return fmt.Errorf("kinesis client error: %w", err)
	}

	// Initialize the LCT stream publisher
	service.Logger.Debug("Initializing LCT publisher", zap.String("LctStreamName", service.cfg.LctStreamName))
	service.lctProducer, err = lct.NewLctProducerFromConfig(&producer.Config{
		StreamName: service.cfg.LctStreamName, Logger: lct.NewProducerLogger(service.Logger), Client: service.kinesisClient,
		GetShards: producer.GetKinesisShardsFunc(service.kinesisClient, service.cfg.LctStreamName),
	}, service.metricsClient)
	if err != nil {
		return fmt.Errorf("lct producer error: %w", err)
	}

	service.hintsClient = hinting.NewHintsClientFromConfig(service.cfg)

	checker := streamchecker.NewStreamCheckerWithTimeout(service.cfg.LctStreamName, service.kinesisClient, 5*time.Second)

	err = checker.DoesStreamExist(context.Background())
	if err != nil {
		return err
	}

	service.lctProducer.Start()

	service.HttpServer, err = NewHttpServer(service)
	if err != nil {
		return err
	}
	atomic.StoreInt32(service.alive, 1)
	atomic.StoreInt32(service.Ready, 1)
	logger.BaseLogger().Info(fmt.Sprintf("Golang runtime version: %v", runtime.Version()))
	logger.BaseLogger().Info(fmt.Sprintf("hello for environment '%s' listening at '%s'\n", utils.FlagEmpty(service.cfg.CbdEnvName), utils.FlagEmpty(service.cfg.ServerAddr)))
	logger.BaseLogger().Info(fmt.Sprintf("Log level: %v", logLevel))
	return service.runUntilToldToShutdown()
}

func hookSignals() <-chan os.Signal {
	quit := make(chan os.Signal, 100)
	signal.Notify(quit, syscall.SIGINT, syscall.SIGTERM)
	return quit
}

func (service *HelloService) runUntilToldToShutdown() error {
	<-service.signals
	return service.shutdown()
}

func panicLogger(desc string) {
	if r := recover(); r != nil {
		logger.BaseLogger().Error(desc, zap.Error(fmt.Errorf("%v", r)))
	}
}

func (service *HelloService) closeHttpServer() error {
	defer panicLogger("Panic closing httpserver")
	if service.HttpServer != nil {
		return service.HttpServer.Shutdown()
	}
	return nil
}

func (service *HelloService) shutdown() error {
	logger.BaseLogger().Info("Shutting down server...")

	httpCloseErr := service.closeHttpServer()

	if service.metricsClient != nil {
		service.metricsClient.CloseConnection()
	}

	service.lctProducer.Stop()

	errors := make([]error, 3)
	errCount := 0
	if httpCloseErr != nil {
		errCount++
		errors[0] = fmt.Errorf("error closing http server %w: ", httpCloseErr)
	}

	loggerStopError := logger.S3Logger().Stop()
	if loggerStopError != nil {
		errCount++
		errors[errCount] = loggerStopError
	}

	if errCount > 0 {
		errStrings := make([]string, errCount)
		errStringsIndex := 0
		for _, err := range errors {
			if err != nil {
				errStrings[errStringsIndex] = err.Error()
				errStringsIndex++
			}
		}
		return fmt.Errorf("shutdown encountered error(s): \n%v", strings.Join(errStrings, "\n"))
	}

	return nil
}

// RunForever -- Main loop of the service
func (service *HelloService) RunForever() {
	logger.BaseLogger().Info(utils.HeaderStr("Hello Service Up and Running"))

	defer func() {
		err := recover()
		if err != nil {
			logger.BaseLogger().Error("HelloService exiting from error", zap.Any("error", err), zap.String("stacktrace", string(debug.Stack())))
		}
	}()

	if err := service.setupAndRun(); err != nil {
		logger.BaseLogger().Error("Service exited with error", zap.Any("error", err))
	}

	logger.BaseLogger().Info("Service exited gracefully")
}
