// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package controllers

import (
	"bytes"
	"encoding/base64"
	"encoding/hex"
	"fmt"
	"hello-api/internal/metrics"
	"net/http"
	"net/http/httptest"
	"strings"
	"testing"
	"time"

	"hello-api/internal/config"
	"hello-api/internal/hinting"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"
	"github.com/felixge/httpsnoop"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"
	"google.golang.org/protobuf/proto"

	"hello-api/internal/app/middleware"
	"hello-api/internal/app/parsers/deviceresponseheader"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

type mockRouterGroup struct {
	routesToHandlers map[string][]gin.HandlerFunc
}

func (mrg *mockRouterGroup) POST(prefix string, handlers ...gin.HandlerFunc) gin.IRoutes {
	mrg.routesToHandlers[prefix] = handlers
	return nil
}

func REDMetricsMock(expectedStatus int) gin.HandlerFunc {
	handler := func(context *gin.Context) {
		func() {
			m := httpsnoop.CaptureMetrics(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
				context.Next()
			}), context.Writer, context.Request)
			path := context.Request.URL.Path

			if strings.Contains(path, "health_check") {
				return
			}

			Expect(m.Duration.Seconds()).ToNot(BeNil())
			Expect(context.Writer.Status()).To(Equal(expectedStatus))
		}()
	}
	return handler
}

var _ = Describe("hello controller", func() {
	testContextHeader := &deviceauth.Context{
		Device: &deviceauth.Device{
			Id: 1337, DeviceType: deviceauth.DeviceType_LINUX,
			DeploymentType: deviceauth.DeploymentType_ENDPOINT, OrgKey: "foobar",
		},
		DeviceGroup:  &deviceauth.DeviceGroup{Id: 1337, OrgId: 1337},
		Organization: &deviceauth.Organization{OrgKey: "foobar", Id: 1337},
	}

	marshalledHeader, _ := proto.Marshal(testContextHeader)
	encodedDeviceContextHeader := base64.StdEncoding.EncodeToString(marshalledHeader)

	helloRequestHeaderMsgPack, _ := hex.DecodeString("c011c0c0c0c0c0c0c0c0c0d30000017005c5960081b14c6173744c697665517565727954696d65cf0000016d15e51e00")

	metricsClient, _ := metrics.NewClient(&config.Config{
		WavefrontEnabled:  true,
		PrometheusEnabled: false,
	})

	Context("dereg handler", func() {
	})

	Context("Hello controller", func() {
		Context("normal hello calls", func() {
			It("Ensure that the correct status code is passed in the middleware chain when context is aborted in one of the handlers", func() {
				helloController := NewHelloController(&config.Config{DeregAuditModeDisabled: false}, nil)
				recorder := httptest.NewRecorder()
				_, testEngine := gin.CreateTestContext(recorder)
				hintResult := hints.NewHintResult(hinttype.HASH_LIST_REFRESH)
				hintResult.LastUpdateTime = timePointerUnixMilli(1337)
				hintResult.LastLiveQueryTime = timePointerUnixMilli(777)
				testEngine.Use(func(c *gin.Context) {
					c.Set(middleware.HintsName, hintResult)
					c.Set(middleware.VersionName, 17)
				}).Use(REDMetricsMock(http.StatusInternalServerError))

				helloController.AddToRouter(testEngine.Group("/services"), func(c *gin.Context) {
					c.Next()
				}, func(c *gin.Context) {
					c.AbortWithStatus(http.StatusInternalServerError)
					c.Next()
				})

				request, _ := http.NewRequest("POST", "/services/hello", bytes.NewBuffer(helloRequestHeaderMsgPack))
				request.Header.Set("Device-Context", encodedDeviceContextHeader)
				testEngine.ServeHTTP(recorder, request)
				Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
			})

			It("returns 200 on hello calls that at well formed", func() {
				helloController := NewHelloController(&config.Config{DeregAuditModeDisabled: false}, nil)
				recorder := httptest.NewRecorder()
				_, testEngine := gin.CreateTestContext(recorder)
				hintResult := hints.NewHintResult(hinttype.HASH_LIST_REFRESH)
				hintResult.LastUpdateTime = timePointerUnixMilli(1337)
				hintResult.LastLiveQueryTime = timePointerUnixMilli(777)
				testEngine.Use(func(c *gin.Context) {
					c.Set(middleware.HintsName, hintResult)
					c.Set(middleware.VersionName, 17)
				})
				helloController.AddToRouter(testEngine.Group("/services"), func(c *gin.Context) {
					c.Next()
				}, func(c *gin.Context) {
					c.Next()
				})

				request, _ := http.NewRequest("POST", "/services/hello", bytes.NewBuffer(helloRequestHeaderMsgPack))
				request.Header.Set("Device-Context", encodedDeviceContextHeader)
				testEngine.ServeHTTP(recorder, request)
				Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
				decoder := deviceresponseheader.NewResponseHeaderDecoder()
				responseBody := recorder.Body.String()
				decodedBody, err := decoder.Decode(responseBody)
				Expect(err).To(BeNil())
				Expect(decodedBody.ErrorCode).To(BeEquivalentTo(0))
				Expect(decodedBody.Message).To(BeEquivalentTo("Success."))
				Expect(decodedBody.Hints).To(BeComparableTo([]int{int(hinttype.HASH_LIST_REFRESH)}))
				Expect(decodedBody.LastUpdateTime).To(BeComparableTo(hintResult.LastUpdateTime))
				Expect(decodedBody.LastLiveQueryTime).To(BeComparableTo(hintResult.LastLiveQueryTime))
				Expect(decodedBody.Version).To(BeComparableTo(int64(17)))
			})

			badHeaderValues := []string{"potato", "shoe", "blueberry", "5", ""}
			for _, badValue := range badHeaderValues {
				It(fmt.Sprintf("ignores erroneous dereg header value of %s and gives good response to sensor", badValue), func() {
					helloController := NewHelloController(&config.Config{DeregAuditModeDisabled: true}, nil)

					hintResult := hints.NewHintResult(hinttype.HASH_LIST_REFRESH)
					hintResult.LastUpdateTime = timePointerUnixMilli(1337)
					hintResult.LastLiveQueryTime = timePointerUnixMilli(777)
					recorder := httptest.NewRecorder()
					_, testEngine := gin.CreateTestContext(recorder)
					testEngine.Use(func(c *gin.Context) {
						c.Set(middleware.HintsName, hintResult)
						c.Set(middleware.VersionName, 17)
					})
					helloController.AddToRouter(testEngine.Group("/services"), func(c *gin.Context) {
						c.Next()
					}, func(c *gin.Context) {
						c.Next()
					})

					request, _ := http.NewRequest("POST", "/services/hello", bytes.NewBuffer(helloRequestHeaderMsgPack))
					request.Header.Set("Device-Context", encodedDeviceContextHeader)

					request.Header.Set("Device-Unauthorized", badValue)
					testEngine.ServeHTTP(recorder, request)
					Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
					decoder := deviceresponseheader.NewResponseHeaderDecoder()
					responseBody := recorder.Body.String()
					decodedBody, err := decoder.Decode(responseBody)
					Expect(err).To(BeNil())
					Expect(decodedBody.ErrorCode).To(BeEquivalentTo(0))
					Expect(decodedBody.Message).To(BeEquivalentTo("Success."))
					Expect(decodedBody.Hints).To(BeComparableTo([]int{int(hinttype.HASH_LIST_REFRESH)}))
					Expect(decodedBody.LastUpdateTime).To(BeComparableTo(hintResult.LastUpdateTime))
					Expect(decodedBody.LastLiveQueryTime).To(BeComparableTo(hintResult.LastLiveQueryTime))
					Expect(decodedBody.Version).To(BeComparableTo(int64(17)))
				})
			}
		})

		It("adds route with appropriate middleware", func() {
			helloController := NewHelloController(&config.Config{}, nil)
			hintsClient := hinting.NewHintsClient(nil, "")
			hintingMiddleware := middleware.HintingMiddleware(hintsClient, metricsClient)
			mockRouter := mockRouterGroup{make(map[string][]gin.HandlerFunc)}
			lctMiddleware := middleware.LctPublisherMiddleware(nil, metricsClient)
			helloController.AddToRouter(&mockRouter, lctMiddleware, hintingMiddleware)
			//goland:noinspection GoPrintFunctions
			helloHandlers := mockRouter.routesToHandlers["/hello"]
			helloHandlerCount := len(helloHandlers)
			Expect(helloHandlerCount).To(BeEquivalentTo(7))
		})

		Context("handles dereg correctly", func() {
			It("returns 200 on dereg when not auditing", func() {
				helloController := NewHelloController(&config.Config{DeregAuditModeDisabled: true}, nil)
				recorder := httptest.NewRecorder()
				_, testEngine := gin.CreateTestContext(recorder)

				helloController.AddToRouter(testEngine.Group("/services"), func(c *gin.Context) {
					c.Next()
				}, func(c *gin.Context) {
					c.Next()
				})
				request, _ := http.NewRequest("POST", "/services/hello", bytes.NewBuffer(helloRequestHeaderMsgPack))
				request.Header.Set("Device-Context", encodedDeviceContextHeader)
				request.Header.Set("Device-Unauthorized", "1")
				testEngine.ServeHTTP(recorder, request)
				Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
				decoder := deviceresponseheader.NewResponseHeaderDecoder()
				responseBody := recorder.Body.String()
				decodedBody, err := decoder.Decode(responseBody)
				Expect(err).To(BeNil())
				Expect(decodedBody.ErrorCode).To(BeEquivalentTo(DeregisteredDeviceErrorCode))
				Expect(decodedBody.Message).To(BeEquivalentTo(DeregisteredDeviceMessage))
			})

			It("returns 503 on dereg when auditing", func() {
				helloController := NewHelloController(&config.Config{DeregAuditModeDisabled: false}, nil)
				recorder := httptest.NewRecorder()
				_, testEngine := gin.CreateTestContext(recorder)
				helloController.AddToRouter(testEngine.Group("/services"), func(c *gin.Context) {
					c.Next()
				}, func(c *gin.Context) {
					c.Next()
				})
				request, _ := http.NewRequest("POST", "/services/hello", bytes.NewBuffer(helloRequestHeaderMsgPack))
				request.Header.Set("Device-Context", encodedDeviceContextHeader)
				request.Header.Set("Device-Unauthorized", "1")
				testEngine.ServeHTTP(recorder, request)
				Expect(recorder.Code).To(BeComparableTo(http.StatusServiceUnavailable))
				decoder := deviceresponseheader.NewResponseHeaderDecoder()
				responseBody := recorder.Body.String()
				_, err := decoder.Decode(responseBody)
				Expect(err).ToNot(BeNil())
			})
		})

		Context("handles pregeg corrrectly", func() {
			It("returns 200 on prereg", func() {
				helloController := NewHelloController(&config.Config{DeregAuditModeDisabled: true}, nil)
				recorder := httptest.NewRecorder()
				_, testEngine := gin.CreateTestContext(recorder)

				helloController.AddToRouter(testEngine.Group("/services"), func(c *gin.Context) {
					c.Next()
				}, func(c *gin.Context) {
					c.Next()
				})
				request, _ := http.NewRequest("POST", "/services/hello", bytes.NewBuffer(helloRequestHeaderMsgPack))
				request.Header.Set("Device-Unauthorized", "2")
				testEngine.ServeHTTP(recorder, request)
				Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
				decoder := deviceresponseheader.NewResponseHeaderDecoder()
				responseBody := recorder.Body.String()
				decodedBody, err := decoder.Decode(responseBody)
				Expect(err).To(BeNil())
				Expect(decodedBody.ErrorCode).To(BeEquivalentTo(backendMaintenanceModeErrroCode))
				Expect(decodedBody.Message).To(BeEquivalentTo(maintenanceModeMessage))
			})
		})
	})
})

func TestController(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Hello controller tests")
}

func timePointerUnixMilli(millis int64) *time.Time {
	t := time.UnixMilli(millis)
	return &t
}
