package controllers

import (
	"fmt"
	"net/http"
	"strconv"

	"hello-api/internal/app/middleware"
	"hello-api/internal/app/parsers/deviceresponseheader"
	"hello-api/internal/config"
	"hello-api/internal/logger"
	"hello-api/internal/metrics"
	"hello-api/internal/utils"

	"github.com/gin-gonic/gin"
	"go.uber.org/zap"
)

var DeviceUnauthorizedHeader string = "Device-Unauthorized"

const (
	DeviceUnauthorizedHeaderValue = "1"
	DeregisteredDeviceMessage     = "Invalid request"
	DeregisteredDeviceErrorCode   = 401
)

type deregHandler struct {
	metricsClient     metrics.Service
	auditModeDisabled bool
}

func newDeregHandler(conf *config.Config, metricsClient metrics.Service) gin.HandlerFunc {
	handler := deregHandler{auditModeDisabled: conf.DeregAuditModeDisabled, metricsClient: metricsClient}
	return handler.doDeregistration
}

func s3LogIfNotSeen(c *gin.Context) {
	deviceContext, err := utils.GetDeviceAuthContext(c)
	if err != nil {
		logger.RequestLoggerFromContext(c).Error("couldn't find device context in dereg middleware", zap.Error(err))
		return
	}
	if deviceContext.Device == nil {
		logger.RequestLoggerFromContext(c).Error("device was nil in deviceAuthContext in dereg middleware")
		return
	}
	deviceId := deviceContext.Device.Id
	orgKey := deviceContext.Device.OrgKey
	deviceAndOrgKey := fmt.Sprintf("%d%s", deviceId, orgKey)
	logger.S3Logger().IfNotSeen([]byte(deviceAndOrgKey)).Info("deregistration", utils.GetAdditionalLogFields(c)...)
}

func deregLoggingAndMetrics(metricsClient metrics.Service, c *gin.Context, auditMode bool) {
	verb := "auditing"
	if !auditMode {
		verb = "handling"
	}
	logger.BaseLogger().Debug(fmt.Sprintf("%s a sensor deregistration", verb))
	utils.AddAdditionalLogFields(c, zap.Bool("audit", auditMode))
	s3LogIfNotSeen(c)
	if metricsClient == nil {
		return
	}
	metricsClient.Count("dereg", map[string]string{"audit": strconv.FormatBool(auditMode)}, 1)
}

func writeDeregResponse(c *gin.Context) error {
	version := c.GetInt(middleware.VersionName)

	header := deviceresponseheader.NewDeviceResponseHeader(int64(version), deviceresponseheader.HelloResponseCode, false,
		DeregisteredDeviceErrorCode, DeregisteredDeviceMessage, nil, nil, nil)
	body, err := deviceresponseheader.NewResponseHeaderEncoder().Encode(header)
	if err != nil {
		return fmt.Errorf("failed to encode response body: %w", err)
	}
	_, err = c.Writer.Write([]byte(body))
	if err != nil {
		return fmt.Errorf("failed to write response body: %w", err)
	}
	return nil
}

func (deregHandler deregHandler) deregOrAudit(c *gin.Context) {
	auditMode := !deregHandler.auditModeDisabled

	deregLoggingAndMetrics(deregHandler.metricsClient, c, auditMode)

	if auditMode {
		c.AbortWithStatus(http.StatusServiceUnavailable)
		return
	}

	err := writeDeregResponse(c)
	if err != nil {
		_ = c.AbortWithError(http.StatusInternalServerError, err)
		return
	}

	c.AbortWithStatus(http.StatusOK)
}

func (deregHandler deregHandler) doDeregistration(c *gin.Context) {
	if !isDeregRequest(c) {
		c.Next()
		return
	}
	deregHandler.deregOrAudit(c)
	c.Abort()
}

func isDeregRequest(c *gin.Context) bool {
	isDereg := c.GetHeader(DeviceUnauthorizedHeader) == DeviceUnauthorizedHeaderValue
	utils.AddAdditionalLogFields(c, zap.Bool("dereg", isDereg))
	utils.SetShouldLog(c, isDereg)
	return isDereg
}
