package controllers

import (
	"net/http"
	"net/http/httptest"

	"hello-api/internal/app/parsers/deviceresponseheader"
	"hello-api/internal/config"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

var _ = Describe("dereg handler", func() {
	Context("bad header values", func() {
		It("Does not dereg on bad header values", func() {
			helloHandler := newDeregHandler(&config.Config{DeregAuditModeDisabled: false}, nil)
			recorder := httptest.NewRecorder()
			testContext, _ := gin.CreateTestContext(recorder)
			testContext.Request, _ = http.NewRequest("POST", "/hello", http.NoBody)
			for _, badHeaderValue := range []string{"potato", "mopar", "chevrolet", "", "2", "5", "13abd"} {
				testContext.Request.Header.Set(DeviceUnauthorizedHeader, badHeaderValue)
				helloHandler(testContext)
				Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
				decoder := deviceresponseheader.NewResponseHeaderDecoder()
				_, err := decoder.Decode(recorder.Body.String())
				Expect(err).ToNot(BeNil())
			}
		})
	})
	Context("audit mode", func() {
		It("dereg handler returns 503 on dereg when auditing", func() {
			helloHandler := newDeregHandler(&config.Config{DeregAuditModeDisabled: false}, nil)
			recorder := httptest.NewRecorder()
			testContext, _ := gin.CreateTestContext(recorder)
			testContext.Request, _ = http.NewRequest("POST", "/hello", http.NoBody)
			testContext.Request.Header.Set(DeviceUnauthorizedHeader, DeviceUnauthorizedHeaderValue)
			helloHandler(testContext)
			Expect(recorder.Code).To(BeComparableTo(http.StatusServiceUnavailable))
		})
	})
	Context("non-audit mode", func() {
		It("returns 200 on dereg when not auditing", func() {
			helloHandler := newDeregHandler(&config.Config{DeregAuditModeDisabled: true}, nil)
			recorder := httptest.NewRecorder()
			testContext, _ := gin.CreateTestContext(recorder)
			testContext.Request, _ = http.NewRequest("POST", "/hello", http.NoBody)
			testContext.Request.Header.Set(DeviceUnauthorizedHeader, DeviceUnauthorizedHeaderValue)
			helloHandler(testContext)
			Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
			decoder := deviceresponseheader.NewResponseHeaderDecoder()
			decodedBody, err := decoder.Decode(recorder.Body.String())
			Expect(err).To(BeNil())
			Expect(decodedBody.ErrorCode).To(BeEquivalentTo(DeregisteredDeviceErrorCode))
			Expect(decodedBody.Message).To(BeEquivalentTo(DeregisteredDeviceMessage))
		})
	})
})
