package controllers

import (
	"errors"
	"net/http"
	"net/http/httptest"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	"hello-api/internal/app/middleware"
	"hello-api/internal/app/parsers/deviceresponseheader"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

type errorEncoder struct{}

func (er *errorEncoder) Encode(_ *deviceresponseheader.DeviceResponseHeader) (string, error) {
	return "", errors.New("error")
}

var _ = Describe("hello handler", func() {
	It("returns 200 code on valid inputs", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		hintResult := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
		version := 18
		lastUpdateTime := time.UnixMilli(1675397106000)    // Friday, February 3, 2023 4:05:06
		lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
		hintResult.LastUpdateTime = &lastUpdateTime
		hintResult.LastLiveQueryTime = &lastLiveQueryTime
		testContext.Set(middleware.HintsName, hintResult)
		testContext.Set(middleware.VersionName, version)
		newHelloHandler(deviceresponseheader.NewResponseHeaderEncoder())(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		decoder := deviceresponseheader.NewResponseHeaderDecoder()
		responseBodyParsed, err := decoder.Decode(recorder.Body.String())
		Expect(err).To(BeNil())
		Expect(responseBodyParsed).To(BeEquivalentTo(deviceresponseheader.NewDeviceResponseHeader(int64(version), HelloResponseCode, true, 0, "Success.", hintResult.Hints.ToIntArray(), &lastUpdateTime, &lastLiveQueryTime)))
	})

	It("returns 500 when encoding the response fails", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/hello", http.NoBody)
		newHelloHandler(&errorEncoder{})(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("returns 500 when LastQueryTime result is missing", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		hints := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
		testContext.Set(middleware.HintsName, hints)
		lastUpdateTime := time.UnixMilli(1675397106000) // Friday, February 3, 2023 4:05:06
		testContext.Set("LastUpdateTime", &lastUpdateTime)

		testContext.Request, _ = http.NewRequest("GET", "/hello", http.NoBody)
		newHelloHandler(deviceresponseheader.NewResponseHeaderEncoder())(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("returns 500 when LastUpdateTime result is missing", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		hints := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
		testContext.Set(middleware.HintsName, hints)
		lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
		testContext.Set("LastQueryTime", &lastLiveQueryTime)

		testContext.Request, _ = http.NewRequest("GET", "/hello", http.NoBody)
		newHelloHandler(&errorEncoder{})(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})
})
