package controllers

import (
	"fmt"
	"net/http"

	"hello-api/internal/app/middleware"
	"hello-api/internal/app/parsers/deviceresponseheader"
	"hello-api/internal/logger"
	"hello-api/internal/metrics"
	"hello-api/internal/utils"

	"github.com/gin-gonic/gin"
	"go.uber.org/zap"
)

const (
	preregHeaderValue               = "2"
	maintenanceModeMessage          = "Maintenance mode"
	backendMaintenanceModeErrroCode = 2000
)

type preregHandler struct {
	metricsClient metrics.Service
}

func newPreregHandler(metricsClient metrics.Service) gin.HandlerFunc {
	handler := preregHandler{metricsClient: metricsClient}
	return handler.doPreregistration
}

func writePreregResponse(c *gin.Context) error {
	version := c.GetInt(middleware.VersionName)

	header := deviceresponseheader.NewDeviceResponseHeader(int64(version), deviceresponseheader.HelloResponseCode, false,
		backendMaintenanceModeErrroCode, maintenanceModeMessage, nil, nil, nil)
	body, err := deviceresponseheader.NewResponseHeaderEncoder().Encode(header)
	if err != nil {
		return fmt.Errorf("failed to encode response body: %w", err)
	}
	_, err = c.Writer.Write([]byte(body))
	if err != nil {
		return fmt.Errorf("failed to write response body: %w", err)
	}
	return nil
}

func (preregHandler preregHandler) prereg(c *gin.Context) {
	logger.BaseLogger().Debug("handling a sensor preregistration")
	if preregHandler.metricsClient != nil {
		preregHandler.metricsClient.Count("prereg", map[string]string{}, 1)
	}
	err := writePreregResponse(c)
	if err != nil {
		_ = c.AbortWithError(http.StatusInternalServerError, err)
		return
	}

	c.AbortWithStatus(http.StatusOK)
}

func (preregHandler preregHandler) doPreregistration(c *gin.Context) {
	if !isPreregRequest(c) {
		c.Next()
		return
	}
	preregHandler.prereg(c)
	c.Abort()
}

func isPreregRequest(c *gin.Context) bool {
	isPrereg := c.GetHeader(DeviceUnauthorizedHeader) == preregHeaderValue
	utils.AddAdditionalLogFields(c, zap.Bool("prereg", isPrereg))
	utils.SetShouldLog(c, isPrereg)
	return isPrereg
}
