// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"encoding/base64"
	"net/http"
	"net/http/httptest"
	"testing"

	"hello-api/internal/utils"

	deviceauthpb "carbonblack.com/protocols/psc/ems/server/deviceauth"
	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"google.golang.org/protobuf/proto"
)

const orgKey = "7ZP4H5XR" // orgkey = ENV local, OrgId 2023

func TestContextMiddlewares(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Middleware tests")
}

var _ = Describe("Device context middleware", func() {
	middleware := DeviceContextMiddleware

	goodOrg := &deviceauthpb.Organization{
		Id:     2023,
		OrgKey: orgKey,
	}
	goodGroup := &deviceauthpb.DeviceGroup{
		Id:    456,
		OrgId: goodOrg.Id,
	}
	goodDevice := &deviceauthpb.Device{
		Id:             123,
		DeploymentType: deviceauthpb.DeploymentType_ENDPOINT,
		DeviceType:     deviceauthpb.DeviceType_LINUX,
		OrgKey:         goodOrg.OrgKey,
	}

	// ----- Begin Tests ------------------------------

	It("returns HTTP 500 when the context header is not supplied", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("POST", "/hello", http.NoBody)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	// -----

	It("returns 200 when the context header is correctly supplied", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		input := &deviceauthpb.Context{Device: goodDevice, DeviceGroup: goodGroup, Organization: goodOrg}
		marshalled, err := proto.Marshal(input)
		Expect(err).To(BeNil())
		b64Encoded := base64.StdEncoding.EncodeToString(marshalled)
		testContext.Request, _ = http.NewRequest("POST", "/hello", http.NoBody)
		testContext.Request.Header.Set(deviceContextHeader, b64Encoded)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		authContext, err := utils.GetDeviceAuthContext(testContext)
		Expect(err).To(BeNil())
		Expect(authContext).ToNot(BeNil())
	})

	It("returns 200 when the context header is correctly supplied, minimal device info", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)

		minDevice := &deviceauthpb.Device{ // No DeviceType or DeploymentType
			Id:     123,
			OrgKey: goodOrg.OrgKey,
		}

		input := &deviceauthpb.Context{Device: minDevice, DeviceGroup: goodGroup, Organization: goodOrg}
		marshalled, err := proto.Marshal(input)
		Expect(err).To(BeNil())
		b64Encoded := base64.StdEncoding.EncodeToString(marshalled)
		testContext.Request, _ = http.NewRequest("POST", "/hello", http.NoBody)
		testContext.Request.Header.Set(deviceContextHeader, b64Encoded)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		authContext, err := utils.GetDeviceAuthContext(testContext)
		Expect(err).To(BeNil())
		Expect(authContext).ToNot(BeNil())
	})
})
