// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"bytes"
	"encoding/hex"
	"net/http"
	"net/http/httptest"
	"time"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

var _ = Describe("Device request header middleware", func() {
	middleware := DeviceRequestHeaderMiddleware

	// ----- Begin Tests ------------------------------
	// NOTE: buffer encodes created from "tests" defined in `internal/app/parsers/devicerequestheader/device_request_msgpack_header_parser_test.go`

	It("returns 200 when header exists and is complete(llqt is uint64)", func() {
		llqt := time.UnixMilli(1657238400000)
		lut := time.UnixMilli(1672617600000)

		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		decodedBuffer, _ := hex.DecodeString("c011c0c0c0c0c0c0c0c0c0d3000001856fc7240081b14c6173744c697665517565727954696d65cf00000181db1b2c00")
		testContext.Request, _ = http.NewRequest("POST", "/hello", bytes.NewBuffer(decodedBuffer))
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		Expect(testContext.Value(VersionName)).To(BeComparableTo(17))
		Expect(testContext.Value(lastUpdateTimeName)).To(BeComparableTo(&lut))
		Expect(testContext.Value(lastQueryTimeName)).To(BeComparableTo(&llqt))
	})

	It("returns 200 when header exists and is complete(llqt is int64)", func() {
		llqt := time.UnixMilli(1568030400000)
		lut := time.UnixMilli(1580644800000)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		decodedBuffer, _ := hex.DecodeString("c011c0c0c0c0c0c0c0c0c0d30000017005c5960081b14c6173744c697665517565727954696d65d30000016d15e51e00")
		testContext.Request, _ = http.NewRequest("POST", "/hello", bytes.NewBuffer(decodedBuffer))
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		Expect(testContext.Value(VersionName)).To(BeComparableTo(17))
		Expect(testContext.Value(lastUpdateTimeName)).To(BeComparableTo(&lut))
		Expect(testContext.Value(lastQueryTimeName)).To(BeComparableTo(&llqt))
	})

	It("returns 200 when header exists and is incomplete(llqt info map is nil)", func() {
		lut := time.UnixMilli(1580644800000)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		decodedBuffer, _ := hex.DecodeString("c011c0c0c0c0c0c0c0c0c0d30000017005c59600")
		testContext.Request, _ = http.NewRequest("POST", "/hello", bytes.NewBuffer(decodedBuffer))
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		Expect(testContext.Value(VersionName)).To(BeComparableTo(17))
		Expect(testContext.Value(lastUpdateTimeName)).To(BeComparableTo(&lut))
		Expect(testContext.Value(lastQueryTimeName)).To(BeNil())
	})

	It("returns 200 when header exists and is incomplete(lut is nil)", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		decodedBuffer, _ := hex.DecodeString("c011c0c0c0c0c0c0c0c0c0")
		testContext.Request, _ = http.NewRequest("POST", "/hello", bytes.NewBuffer(decodedBuffer))
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		Expect(testContext.Value(VersionName)).To(BeComparableTo(17))
		Expect(testContext.Value(lastUpdateTimeName)).To(BeNil())
	})

	// ----- Negative Tests ------------------------------

	It("fails with HTTP 500 if context header is empty", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("POST", "/hello", http.NoBody)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("fails with HTTP 500 if context header is bogus", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("POST", "/hello", bytes.NewBufferString(""))
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})
})
