// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"errors"
	"net/http"
	"net/http/httptest"
	"time"

	"hello-api/internal/app/parsers/devicecontext"
	"hello-api/internal/config"
	"hello-api/internal/metrics"
	"hello-api/internal/utils"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

type mockHinter struct {
	shouldThrowError   error
	getHintsCalled     bool
	getHintsCalledWith *hints.JsonHintRequest
	hintResult         *hints.HintResult
}

func (mock *mockHinter) GetHints(hintRequest *hints.JsonHintRequest) (*hints.HintResult, error) {
	mock.getHintsCalled = true
	mock.getHintsCalledWith = hintRequest

	return mock.hintResult, mock.shouldThrowError
}

var _ = Describe("hinting middleware", func() {
	testDeviceContext := devicecontext.DeviceAuthContext{Context: &deviceauth.Context{
		Device: &deviceauth.Device{
			OrgKey:         "ORGKEY",
			Id:             1337,
			DeploymentType: deviceauth.DeploymentType_AZURE,
			DeviceType:     deviceauth.DeviceType_MAC,
		},
		Organization: &deviceauth.Organization{OrgKey: "ORGKEY"},
		DeviceGroup:  &deviceauth.DeviceGroup{Id: 1337},
	}}

	metricsClient, _ := metrics.NewClient(&config.Config{
		WavefrontEnabled:  true,
		PrometheusEnabled: false,
	})

	It("fails with HTTP 500 if context is empty", func() {
		hinter := &mockHinter{}
		middleware := HintingMiddleware(hinter, metricsClient)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("fails with HTTP 500 if context is invalid", func() {
		invalidContexts := []*deviceauth.Context{{}, {Device: &deviceauth.Device{Id: 1}}, {Device: &deviceauth.Device{Id: 1}, DeviceGroup: &deviceauth.DeviceGroup{Id: 1}}}
		hinter := &mockHinter{}
		middleware := HintingMiddleware(hinter, metricsClient)
		for _, invalidContext := range invalidContexts {
			recorder := httptest.NewRecorder()
			testContext, _ := gin.CreateTestContext(recorder)
			testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
			utils.SetDeviceAuthContext(testContext, &devicecontext.DeviceAuthContext{Context: invalidContext})
			middleware(testContext)
			Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
		}
	})

	It("returns an empty hint result if there was an error interacting with the hinting microservice", func() {
		hinter := &mockHinter{shouldThrowError: errors.New("SOME HINTING ERROR")}
		middleware := HintingMiddleware(hinter, metricsClient)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
		utils.SetDeviceAuthContext(testContext, &testDeviceContext)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		result, err := utils.ContextGet[*hints.HintResult](testContext, HintsName)
		Expect(err).To(BeNil())
		Expect(result).ToNot(BeNil())
		Expect(result.Hints).To(BeNil())
		Expect(result.LastLiveQueryTime).To(BeNil())
		Expect(result.LastUpdateTime).To(BeNil())
	})

	It("returns 200 when header does exist but lastLiveQueryTime is nil", func() {
		version := 1
		lastUpdateTime := time.UnixMilli(1675397106000) // Friday, February 3, 2023 4:05:06
		// lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
		hinter := &mockHinter{hintResult: &hints.HintResult{}}
		middleware := HintingMiddleware(hinter, metricsClient)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
		utils.SetDeviceAuthContext(testContext, &testDeviceContext)
		testContext.Set(VersionName, version)
		testContext.Set(lastUpdateTimeName, &lastUpdateTime)
		testContext.Set(lastQueryTimeName, nil)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		hint, exists := testContext.Get(HintsName)
		Expect(exists).To(BeTrue())
		hint, ok := hint.(*hints.HintResult)
		Expect(ok).To(BeTrue())
		Expect(hinter.hintResult).To(BeComparableTo(hint))
		Expect(hinter.getHintsCalledWith.Version).To(BeEquivalentTo(&version))
		Expect(*hinter.getHintsCalledWith.LastUpdateTime).To(BeEquivalentTo(lastUpdateTime.UnixMilli()))
		Expect(hinter.getHintsCalledWith.LastLiveQueryTime).To(BeNil())
		Expect(hinter.getHintsCalledWith.DeviceType).To(BeEquivalentTo("MAC"))
		Expect(hinter.getHintsCalledWith.DeploymentType).To(BeEquivalentTo("AZURE"))
		Expect(hinter.getHintsCalledWith.OrgKey).To(BeEquivalentTo("ORGKEY"))
		Expect(hinter.getHintsCalledWith.DeviceId).To(BeEquivalentTo(1337))
		Expect(hinter.getHintsCalledWith.DeviceGroupId).To(BeEquivalentTo(1337))
	})

	It("returns 200 when header does exist", func() {
		version := 1
		lastUpdateTime := time.UnixMilli(1675397106000)    // Friday, February 3, 2023 4:05:06
		lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
		hinter := &mockHinter{hintResult: &hints.HintResult{}}
		middleware := HintingMiddleware(hinter, metricsClient)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
		utils.SetDeviceAuthContext(testContext, &testDeviceContext)
		testContext.Set(VersionName, version)
		testContext.Set(lastUpdateTimeName, &lastUpdateTime)
		testContext.Set(lastQueryTimeName, &lastLiveQueryTime)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		hint, exists := testContext.Get(HintsName)
		Expect(exists).To(BeTrue())
		hint, ok := hint.(*hints.HintResult)
		Expect(ok).To(BeTrue())
		Expect(hinter.hintResult).To(BeComparableTo(hint))
		Expect(hinter.getHintsCalledWith.Version).To(BeEquivalentTo(&version))
		Expect(*hinter.getHintsCalledWith.LastUpdateTime).To(BeEquivalentTo(lastUpdateTime.UnixMilli()))
		Expect(*hinter.getHintsCalledWith.LastLiveQueryTime).To(BeEquivalentTo(lastLiveQueryTime.UnixMilli()))
		Expect(hinter.getHintsCalledWith.DeviceType).To(BeEquivalentTo("MAC"))
		Expect(hinter.getHintsCalledWith.DeploymentType).To(BeEquivalentTo("AZURE"))
		Expect(hinter.getHintsCalledWith.OrgKey).To(BeEquivalentTo("ORGKEY"))
		Expect(hinter.getHintsCalledWith.DeviceId).To(BeEquivalentTo(1337))
		Expect(hinter.getHintsCalledWith.DeviceGroupId).To(BeEquivalentTo(1337))
	})
})
