// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"errors"
	"net/http"
	"net/http/httptest"

	"hello-api/internal/app/parsers/devicecontext"
	"hello-api/internal/config"
	"hello-api/internal/metrics"
	"hello-api/internal/utils"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"carbonblack.com/protocols/psc/ems/server/deviceauth"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

type mockLctPublisher struct {
	shouldThrowError  error
	publishCalled     bool
	publishCalledWith *devicelastcontacttime.DeviceLastContactTime
}

func (mock *mockLctPublisher) Publish(lct *devicelastcontacttime.DeviceLastContactTime) error {
	mock.publishCalled = true
	mock.publishCalledWith = lct
	return mock.shouldThrowError
}

var _ = Describe("lct publishing middleware", func() {
	metricsClient, _ := metrics.NewClient(&config.Config{
		WavefrontEnabled:  true,
		PrometheusEnabled: false,
	})

	It("fails with HTTP 500 if context is empty", func() {
		publisher := &mockLctPublisher{}
		middleware := LctPublisherMiddleware(publisher, metricsClient)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("fails with HTTP 500 if context is not empty but a failure to publish to stream occurs", func() {
		publisher := &mockLctPublisher{shouldThrowError: errors.New("SOME PUBLISHING ERROR")}
		middleware := LctPublisherMiddleware(publisher, metricsClient)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
		utils.SetDeviceAuthContext(testContext, &devicecontext.DeviceAuthContext{Context: &deviceauth.Context{
			Device: &deviceauth.Device{
				OrgKey: "ORGKEY",
				Id:     1337,
			},
		}})
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("returns 200 when header does exist", func() {
		publisher := &mockLctPublisher{}
		middleware := LctPublisherMiddleware(publisher, metricsClient)
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/test", http.NoBody)
		utils.SetDeviceAuthContext(testContext, &devicecontext.DeviceAuthContext{Context: &deviceauth.Context{
			Device: &deviceauth.Device{
				OrgKey: "ORGKEY",
				Id:     1337,
			},
		}})
		middleware(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))
		Expect(publisher.publishCalled).To(BeComparableTo(true))
		Expect(publisher.publishCalledWith.DeviceId).To(BeComparableTo(int64(1337)))
		Expect(publisher.publishCalledWith.OrgKey).To(BeComparableTo("ORGKEY"))
		Expect(publisher.publishCalledWith.DeregisterAfterTime).To(BeComparableTo(int64(0)))
		Expect(publisher.publishCalledWith.LastContactTimestamp).ToNot(BeComparableTo(int64(0)))
	})
})
