package middleware

import (
	"net/http"
	"strconv"
	"strings"

	"hello-api/internal/logger"

	"github.com/gin-gonic/gin"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"

	"github.com/felixge/httpsnoop"
)

const (
	// Common Labels
	labelMethod        = "method"
	labelCode          = "code"
	labelOperationName = "operation_name"

	// Requests Total Metric Name and Labels
	metricNameRequestsTotal = "requests_total"

	// Request Duration Metric Name and Labels
	metricNameRequestsDurationSeconds      = "request_duration_seconds"
	metricNameRequestsDurationSecondsGauge = "request_duration_seconds_gauge"
)

type Metrics struct {
	requestsReceived     *prometheus.CounterVec
	requestDuration      *prometheus.HistogramVec
	requestDurationGauge *prometheus.GaugeVec
}

func RegisterPromMetrics() *Metrics {
	logger.BaseLogger().Debug("-------- ATTEMPTING TO REGISTER METRICS -----------")
	reqReceived := prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Namespace: "http",
			Name:      metricNameRequestsTotal,
			Help:      "Total number of requests received.",
		},
		[]string{labelOperationName, labelMethod, labelCode},
	)
	prometheus.MustRegister(reqReceived)

	reqDuration := prometheus.NewHistogramVec(
		prometheus.HistogramOpts{
			Namespace: "http",
			Name:      metricNameRequestsDurationSeconds,
			Help:      "Duration of a request in seconds.",
		},
		[]string{labelOperationName, labelMethod, labelCode},
	)
	prometheus.MustRegister(reqDuration)

	// Because APM doesn't support histograms, we will also track request duration as a gauge
	reqDurationGauge := prometheus.NewGaugeVec(prometheus.GaugeOpts{
		Namespace: "http",
		Name:      metricNameRequestsDurationSecondsGauge,
		Help:      "Duration of a request in seconds, as a gauge",
	},
		[]string{labelOperationName, labelMethod, labelCode},
	)
	prometheus.MustRegister(reqDurationGauge)

	return &Metrics{
		requestsReceived:     reqReceived,
		requestDuration:      reqDuration,
		requestDurationGauge: reqDurationGauge,
	}
}

func (m *Metrics) Default() http.Handler {
	return promhttp.Handler()
}

func (m *Metrics) IncRequests(operationName string, method string, code int) {
	m.requestsReceived.WithLabelValues(operationName, method, strconv.Itoa(code)).Inc()
}

func (m *Metrics) ObsDuration(operationName string, method string, code int, duration float64) {
	m.requestDuration.WithLabelValues(operationName, method, strconv.Itoa(code)).Observe(duration)
	m.requestDurationGauge.WithLabelValues(operationName, method, strconv.Itoa(code)).Set(duration)
}

func REDMetrics(promMetrics *Metrics) gin.HandlerFunc {
	handler := func(context *gin.Context) {
		func() {
			m := httpsnoop.CaptureMetrics(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
				context.Next()
			}), context.Writer, context.Request)
			path := context.Request.URL.Path

			if strings.Contains(path, "health_check") {
				return
			}

			promMetrics.IncRequests(operationNameFunc(context.Request), context.Request.Method, context.Writer.Status())
			promMetrics.ObsDuration(operationNameFunc(context.Request), context.Request.Method, context.Writer.Status(), m.Duration.Seconds())
		}()
	}
	return handler
}

const httpText = "HTTP "

func operationNameFunc(r *http.Request) string {
	// quick and dirty url bining
	splitPath := strings.Split(r.URL.Path, "/")
	if r.Method == "DELETE" {
		if len(splitPath) < 2 {
			// unknown URL format
			return httpText + r.Method
		}
		operationName := splitPath[len(splitPath)-2]
		return httpText + r.Method + " " + operationName
	}
	if len(splitPath) < 1 {
		// unknown URL format
		return httpText + r.Method
	}
	operationName := splitPath[len(splitPath)-1]
	return httpText + r.Method + " " + operationName
}
