// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"net/http"
	"strings"
	"time"

	"hello-api/internal/logger"
	"hello-api/internal/utils"

	"github.com/gin-gonic/gin"
	"github.com/google/uuid"
	"github.com/opentracing-contrib/go-gin/ginhttp"
	"github.com/opentracing/opentracing-go"
	"github.com/opentracing/opentracing-go/ext"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
)

// Tracing -- Handles middleware tracing
func (mwf *MiddlewareUtils) Tracing() gin.HandlerFunc {
	return ginhttp.Middleware(mwf.metricsClient.Tracer(), mwf.tracingSpanObserverOption(), mwf.tracingOperationNameOption())
}

func (mwf *MiddlewareUtils) GetTracingChain() gin.HandlersChain {
	if mwf.metricsClient != nil && mwf.metricsClient.Tracer() != nil {
		return gin.HandlersChain{
			mwf.Tracing(),
			mwf.GetSpanErrorTaggingMiddleware,
			RecoveryMiddleware,
			mwf.Logging,
			mwf.TracingRequestIdAndLogger,
			TimeoutMiddleware(mwf.cfg.HttpHandleTimeout),
		}
	}
	return gin.HandlersChain{
		RecoveryMiddleware,
		mwf.Logging,
		mwf.TracingRequestIdAndLogger,
		TimeoutMiddleware(mwf.cfg.HttpHandleTimeout),
	}
}

func (mwf *MiddlewareUtils) Logging(c *gin.Context) {
	defer func(start time.Time) {
		latency := time.Since(start)

		// Fill in additional information seen in the transaction
		logFields := []zap.Field{
			zap.String("remote-address", c.Request.RemoteAddr),
			zap.String("method", c.Request.Method),
			zap.String("request-uri", c.Request.RequestURI),
			zap.String("accept", strings.Join(c.Request.Header["Accept"], " ")),
			zap.String("proto", c.Request.Proto),
			zap.Int("status", c.Writer.Status()),
			zap.Int("length", int(c.Request.ContentLength)),
			zap.String("user-agent", c.Request.UserAgent()),
			zap.Duration("latency", latency),
			zap.Strings("errors", c.Errors.Errors()),
		}

		logFields = append(logFields, utils.GetAdditionalLogFields(c)...)
		requestLogger := logger.RequestLoggerFromContext(c.Request.Context())

		if logger.LogLevel().Enabled(zapcore.DebugLevel) {
			logFields = append(logFields, utils.GetAdditionalDebugLogFields(c)...)
		}

		logMsg := ""
		logFunc := requestLogger.Debug
		switch {
		case c.Writer.Status() == http.StatusInternalServerError:
			logFunc = requestLogger.Error
			logMsg = "Server Error"
		case utils.GetShouldLog(c):
			logFunc = requestLogger.Info
		default:
		}
		logFunc(logMsg, logFields...)
	}(time.Now())

	c.Next()
}

func (mwf *MiddlewareUtils) TracingRequestIdAndLogger(c *gin.Context) {
	var err error
	if sp := opentracing.SpanFromContext(c.Request.Context()); sp != nil {
		err = opentracing.GlobalTracer().Inject(
			sp.Context(),
			opentracing.HTTPHeaders,
			opentracing.HTTPHeadersCarrier(c.Writer.Header()),
		)
	}
	requestID := c.Writer.Header().Get(mwf.cfg.RequestIDHeader)
	if requestID == "" {
		requestID = uuid.New().String() // New request?  Need an ID
	}
	c.Writer.Header().Set(mwf.cfg.RequestIDHeader, requestID)
	l := mwf.RequestLogger.With(zap.String("request-id", requestID))
	if err != nil {
		l.Info(err.Error())
	}
	*c.Request = *c.Request.Clone(logger.ContextWithRequestLogger(c.Request.Context(), l))
	c.Next()
}

func (mwf *MiddlewareUtils) tracingOperationNameOption() ginhttp.MWOption {
	return ginhttp.OperationNameFunc(func(r *http.Request) string {
		return r.RequestURI
	})
}

func (mwf *MiddlewareUtils) tracingSpanObserverOption() ginhttp.MWOption {
	return ginhttp.MWSpanObserver(func(sp opentracing.Span, r *http.Request) {
		sp.SetTag("http.remote_address", r.RemoteAddr)
		sp.SetTag("http.user_agent", r.UserAgent())
		sp.SetTag("http.accept", r.Header.Get("Accept"))
	})
}

func (mwf *MiddlewareUtils) GetSpanErrorTaggingMiddleware(c *gin.Context) {
	defer func(c *gin.Context) {
		if sp := opentracing.SpanFromContext(c.Request.Context()); sp != nil {
			wasErrorReported := c.GetBool("error") || len(c.Errors) > 0
			sp.SetTag(string(ext.Error), wasErrorReported)
		}
	}(c)
	c.Next()
}
