// Copyright (c) 2023 VMware Carbon Black, all rights reserved

// Package devicecontext handles the parsing of Device Context information stored in the request header
package devicecontext

import (
	"encoding/base64"
	"errors"
	"fmt"
	"strings"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"
	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
	"go.uber.org/zap/zapcore"
	"google.golang.org/protobuf/proto"
)

const DevDashboardDeregister = "DEV_DASHBOARD_DEREGISTER"

type (
	DeviceAuthResponseParser interface {
		ParseResponse(deviceContextHeader string) (*DeviceAuthContext, error)
	}

	deviceAuthResponseParser struct{}

	DeviceAuthContext struct {
		*deviceauth.Context
	}
)

// MarshalLogObject implements zapcore.ObjectMarshaler.
func (d DeviceAuthContext) MarshalLogObject(zapEncoder zapcore.ObjectEncoder) error {
	if d.Device != nil {
		zapEncoder.AddInt64("device-id", d.Device.Id)
		zapEncoder.AddString("org-key", d.Device.OrgKey)
		zapEncoder.AddString("device-type", d.Device.DeviceType.String())
		zapEncoder.AddString("deployment-type", d.Device.DeploymentType.String())
		orgKey, err := orgkeylib.Decode(d.Device.OrgKey)
		if err == nil {
			zapEncoder.AddString("legacy-env", orgKey.Env())
		}

		properties := d.Device.Properties
		version := "not-found"
		if properties != nil {
			version = properties["Scargo Client Version"]
		}
		zapEncoder.AddString("sensor-version", version)
		changeTimestamps := d.Device.ChangeTimestamps

		if changeTimestamps != nil {
			deregTime, found := changeTimestamps[DevDashboardDeregister]
			if found {
				if deregTime != nil {
					zapEncoder.AddTime("dereg_time", deregTime.AsTime())
				} else {
					zapEncoder.AddString("dereg_time", "null")
				}
			} else {
				zapEncoder.AddString("dereg_time", "not_found")
			}
		}
	} else {
		zapEncoder.AddString("device-context-error", "context.Device was nil")
	}
	if d.DeviceGroup != nil {
		zapEncoder.AddInt64("group-id", d.DeviceGroup.Id)
	} else {
		zapEncoder.AddString("device-context-error", "context.DeviceGroup was nil")
	}
	if d.Organization != nil {
		zapEncoder.AddString("org-key", d.Organization.OrgKey)
		zapEncoder.AddInt64("org-id", d.Organization.Id)
	} else {
		zapEncoder.AddString("device-context-error", "context.Organization was nil")
	}
	return nil
}

func NewDeviceAuthResponseParser() DeviceAuthResponseParser {
	return &deviceAuthResponseParser{}
}

// ParseResponse -- Decode the protobuf response returned as a Base64 encoded string
//   - as per 'device_auth.proto'
func (parser *deviceAuthResponseParser) ParseResponse(deviceContextHeader string) (*DeviceAuthContext, error) {
	if strings.TrimSpace(deviceContextHeader) == "" {
		return nil, errors.New("device context header was empty")
	}

	decodedHeader, err := base64.StdEncoding.DecodeString(deviceContextHeader)
	if err != nil {
		return nil, fmt.Errorf("failed to b64 decode device context header %w", err)
	}

	context := new(deviceauth.Context)
	err = proto.Unmarshal(decodedHeader, context)
	if err != nil {
		return nil, fmt.Errorf("failed to unmarshal device context header %w", err)
	}
	deviceAuthContext := DeviceAuthContext{context}

	deviceAuthContext.sanitizeContext()

	return &deviceAuthContext, nil
}

func (ctx *DeviceAuthContext) sanitizeContext() {
	// Remove any default change timestamps
	keysToDelete := []string{}
	if ctx.Device != nil && ctx.Device.ChangeTimestamps != nil {
		for k, val := range ctx.Device.ChangeTimestamps {
			if val.Seconds == 0 && val.Nanos == 0 {
				keysToDelete = append(keysToDelete, k)
			}
		}
		for _, k := range keysToDelete {
			delete(ctx.Device.ChangeTimestamps, k)
		}
	}
}

func (ctx *DeviceAuthContext) Validate() (err error) {
	if ctx.Device == nil {
		return errors.New("no device included in device context header")
	}
	if ctx.DeviceGroup == nil {
		return errors.New("no device group included in device context header")
	}
	if ctx.Organization == nil {
		return errors.New("no org included in device context header")
	}
	return nil
}
