// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package devicecontext

import (
	"encoding/base64"
	"fmt"
	"testing"

	"github.com/golang/protobuf/ptypes/timestamp"

	deviceauthpb "carbonblack.com/protocols/psc/ems/server/deviceauth"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"google.golang.org/protobuf/proto"
)

var t *testing.T

func TestDeviceAuthResponseParsing(test *testing.T) {
	RegisterFailHandler(Fail)
	t = test
	RunSpecs(test, "protobuf device auth response header")
}

func createDeviceContextConsumer() DeviceAuthResponseParser {
	return NewDeviceAuthResponseParser()
}

///// Begin tests ////////////////////////////////////////

var _ = Describe("device context parser", func() {
	ORGKEY := "67ZP4HX7" // Generated OrgKey for ENV 'local', OrgId '3'

	org := &deviceauthpb.Organization{
		Id:     3,
		OrgKey: ORGKEY,
	}

	group := &deviceauthpb.DeviceGroup{
		Id:    2,
		OrgId: org.Id,
	}

	device := &deviceauthpb.Device{
		Id:             1,
		DeploymentType: deviceauthpb.DeploymentType_ENDPOINT,
		DeviceType:     deviceauthpb.DeviceType_LINUX,
		OrgKey:         org.OrgKey,
		ChangeTimestamps: map[string]*timestamp.Timestamp{
			"foo": nil,
			"bar": nil,
			"boo": {
				Seconds: 0,
				Nanos:   0,
			},
			"good": {
				Seconds: 10000,
				Nanos:   10000,
			},
			"good2": {
				Seconds: 200000,
				Nanos:   0,
			},
		},
	}

	parser := NewDeviceAuthResponseParser()
	Context("Testing valid input", func() {
		It("assures encode/decode transparency", func() {
			input := &deviceauthpb.Context{Device: device, DeviceGroup: group, Organization: org}
			marshalled, err := proto.Marshal(input)
			Expect(err).To(BeNil())
			b64Encoded := base64.StdEncoding.EncodeToString(marshalled)
			resp, err2 := parser.ParseResponse(b64Encoded)
			Expect(err2).To(BeNil())
			Expect(resp).ToNot(BeNil())

			// ensure transparency
			Expect(resp.Device.Id).To(BeEquivalentTo(device.Id))
			Expect(resp.Device.DeploymentType).To(BeEquivalentTo(device.DeploymentType))
			Expect(resp.Device.DeviceType).To(BeEquivalentTo(device.DeviceType))
			Expect(resp.Device.OrgKey).To(BeEquivalentTo(device.OrgKey))

			Expect(resp.DeviceGroup.Id).To(BeEquivalentTo(group.Id))
			Expect(resp.DeviceGroup.OrgId).To(BeEquivalentTo(group.OrgId))

			Expect(resp.Organization.Id).To(BeEquivalentTo(org.Id))
			Expect(resp.Organization.OrgKey).To(BeEquivalentTo(org.OrgKey))
		})

		Context("with change timestamps", func() {
			It("removes null and empty change timestamps from the map", func() {
				deviceContextConsumer := createDeviceContextConsumer()

				// Create valid device context
				input := &deviceauthpb.Context{Device: device, DeviceGroup: group, Organization: org}

				marshalled, err := proto.Marshal(input)
				Expect(err).To(BeNil())
				b64Encoded := base64.StdEncoding.EncodeToString(marshalled)

				resp, err := deviceContextConsumer.ParseResponse(b64Encoded)

				Expect(err).To(BeNil())
				Expect(resp).ToNot(BeNil())

				// ensure transparency
				Expect(resp.Device.Id).To(BeEquivalentTo(device.Id))
				Expect(resp.Device.DeploymentType).To(BeEquivalentTo(device.DeploymentType))
				Expect(resp.Device.DeviceType).To(BeEquivalentTo(device.DeviceType))
				Expect(resp.Device.OrgKey).To(BeEquivalentTo(device.OrgKey))

				Expect(resp.DeviceGroup.Id).To(BeEquivalentTo(group.Id))
				Expect(resp.DeviceGroup.OrgId).To(BeEquivalentTo(group.OrgId))

				Expect(resp.Organization.Id).To(BeEquivalentTo(org.Id))
				Expect(resp.Organization.OrgKey).To(BeEquivalentTo(org.OrgKey))

				Expect(resp.Device.ChangeTimestamps).To(HaveLen(2))
				Expect(resp.Device.ChangeTimestamps).To(HaveKey("good"))
				Expect(resp.Device.ChangeTimestamps).To(HaveKey("good2"))
				Expect(resp.Device.ChangeTimestamps).ToNot(HaveKey("foo"))
			})
		})
	})

	Context("Testing invalid inputs", func() {
		It("returns error when given empty input", func() {
			context, err := parser.ParseResponse("")
			Expect(err).ToNot(BeNil())
			Expect(context).To(BeNil())
		})

		It("returns error when given bogus input", func() {
			context, err := parser.ParseResponse("GARBAGE")
			Expect(err).ToNot(BeNil())
			Expect(context).To(BeNil())
		})

		It("returns error with missing device info on validation", func() {
			input := &deviceauthpb.Context{DeviceGroup: group, Organization: org}
			marshalled, err := proto.Marshal(input)
			Expect(err).To(BeNil())
			b64Encoded := base64.StdEncoding.EncodeToString(marshalled)
			authContext, err := parser.ParseResponse(b64Encoded)
			Expect(err).To(BeNil())
			err = authContext.Validate()
			Expect(fmt.Sprintln(err)).To(ContainSubstring("no device included in device context header"))
		})

		It("returns error with data that cannot be unmarshaled", func() {
			_, err := parser.ParseResponse("DEADBEEF")
			Expect(err).ToNot(BeNil())
			Expect(fmt.Sprintln(err)).To(ContainSubstring("failed to unmarshal device context header proto"))
		})

		It("returns error with missing group info", func() {
			input := &deviceauthpb.Context{Device: device, Organization: org}
			marshalled, err := proto.Marshal(input)
			Expect(err).To(BeNil())
			b64Encoded := base64.StdEncoding.EncodeToString(marshalled)
			authContext, err := parser.ParseResponse(b64Encoded)
			Expect(err).To(BeNil())
			err = authContext.Validate()
			Expect(fmt.Sprintln(err)).To(ContainSubstring("no device group included in device context header"))
		})
		It("returns error with missing org info", func() {
			input := &deviceauthpb.Context{Device: device, DeviceGroup: group}
			marshalled, err := proto.Marshal(input)
			Expect(err).To(BeNil())
			b64Encoded := base64.StdEncoding.EncodeToString(marshalled)
			authContext, err := parser.ParseResponse(b64Encoded)
			Expect(err).To(BeNil())
			err = authContext.Validate()
			Expect(err).ToNot(BeNil())
			Expect(fmt.Sprintln(err)).To(ContainSubstring("no org included in device context header"))
		})
	})
})
