// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package devicerequestheader

import (
	"fmt"
	"io"
	"time"

	"github.com/vmihailenco/msgpack/v5"
)

type (
	// RequestHeaderDecoder is the interface to parse msgpack headers
	RequestHeaderDecoder interface {
		ParseMsgPack(io.Reader) (int, *time.Time, *time.Time, error)
	}

	requestHeaderDecoder struct{}
)

// NewRequestHeaderDecoder returns a msgpack RequestHeaderDecoder interface
func NewRequestHeaderDecoder() RequestHeaderDecoder {
	return &requestHeaderDecoder{}
}

func timeIntToTimePointer(timeAsUnixMillis int64) *time.Time {
	t := time.UnixMilli(timeAsUnixMillis)
	return &t
}

// ParseMsgPack takes the msgpack header and returns version, lastUpdateTime, lastLiveQueryTime and an optional error
//
// for device msgpack header format, see:
//   - https://confluence.carbonblack.local/pages/viewpage.action?spaceKey=ENDOC&title=Device+Request+Protocol#DeviceRequestProtocol-4.1.2RequestHeader
func (s *requestHeaderDecoder) ParseMsgPack(reader io.Reader) (version int, lastUpdateTime, lastLiveQueryTime *time.Time, err error) {
	decoder := msgpack.NewDecoder(reader)
	// field 1: TimeSent byte[8]
	if err = decoder.Skip(); err != nil {
		err = fmt.Errorf("failed to skip field 1: %w", err)
		return
	}

	// field 2: Version long
	version, err = decoder.DecodeInt()
	if err != nil {
		err = fmt.Errorf("failed to decode Version %w", err)
		return
	}

	// skip fields 3...11
	for i := 0; i < 9; i++ {
		if skipErr := decoder.Skip(); skipErr != nil {
			return
		}
	}

	// field 12: LastUpdateTime. It's NOT mandatory.
	// We don't return an error if not found in the request msgpack.
	lastUpdateTimeInt, err := decoder.DecodeInt64()
	if err != nil {
		lastUpdateTime = nil
		err = nil
	} else {
		lastUpdateTime = timeIntToTimePointer(lastUpdateTimeInt)
	}

	// field 13: Supported as of V17 as Info field. Info field is NOT mandatory.
	// We don't return an error if not found in the request msgpack.
	if version >= 17 {
		var info map[string]interface{}
		info, err = decoder.DecodeMap()
		if err != nil {
			lastLiveQueryTime = nil
			err = nil
		} else {
			lastLiveQueryTimeRaw := info["LastLiveQueryTime"]
			if lastLiveQueryTimeRaw != nil {
				switch t := lastLiveQueryTimeRaw.(type) {
				case int64:
					lastLiveQueryTime = timeIntToTimePointer(t)
				case uint64:
					lastLiveQueryTime = timeIntToTimePointer(int64(t))
				default:
					lastLiveQueryTime = nil
				}
			}
		}
	}
	return
}
