// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package devicerequestheader

import (
	"bytes"
	"encoding/hex"
	"errors"
	"fmt"
	"io"
	"testing"
	"time"

	"go.uber.org/zap"

	"hello-api/internal/logger"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"github.com/vmihailenco/msgpack/v5"
)

type testData struct {
	version                 int
	lastLiveQueryTimeInt64  *int64
	lastLiveQueryTimeUint64 *uint64
	lastUpdateTime          *int64
	expectErr               error
}

func timePtrToIntPtr(t *time.Time) *int64 {
	if t == nil {
		return nil
	}
	timeInt := t.UnixMilli()
	return &timeInt
}

func timePtrToUIntPtr(t *time.Time) *uint64 {
	if t == nil {
		return nil
	}
	timeInt := uint64(t.UnixMilli())
	return &timeInt
}

// Encode -- Encodes test data info a standard Request Header msgpack object that is then converted to a hex string
//   - see https://confluence.carbonblack.local/display/ENDOC/Device+Request+Protocol#DeviceRequestProtocol-4.1.2RequestHeader
func (td testData) Encode() (encString string, err error) {
	var buf bytes.Buffer
	encoder := msgpack.NewEncoder(&buf)

	// NOTE: Many of these fields are normally mandatory but are skipped for the purposes of unit testing since
	//       they are ignored by the hello service

	// Field 1 -- The time the message was sent to the cloud
	if err = encoder.EncodeNil(); err != nil { // SKIPPED
		return
	}

	// Field 2 -- version of the request
	if err = encoder.EncodeInt(int64(td.version)); err != nil {
		return
	}

	// Field 3 -- Unique id for a device. SHA256 hash of data that uniquely identifies a given device
	// Field 4 -- A string returned in the Registration Response that identifies an administrative domain such as a company
	// Field 5 -- Device Type; Windows=1, Android=2, OSX=3, IOS=4, Linux=5
	// Field 6 -- Code identifying the message type
	// Field 7 -- Sub-code to help discriminate message contents within the message type
	// Field 8 -- Unique message identifier optionally set by the client side to enable request/response correlation in case of asynchronous messaging
	// Field 9 -- Agent Location Values; Unknown=0, On site=1, Off site=2
	// Field 10 -- Local Address; Contains optional name/value pairs
	// Field 11 -- Meant for testing - Boolean indicator to let the backend know that it should not do any processing but merely echo back successful dummy response
	for i := 3; i <= 11; i++ {
		if err = encoder.EncodeNil(); err != nil { // SKIPPED
			return
		}
	}

	// Field 12 -- Timestamp of when the last time the sensor received a hint indicator to pull something new from the cloud
	if td.lastUpdateTime != nil {
		err = encoder.EncodeInt64(*td.lastUpdateTime)
		if err != nil {
			return
		}
	}

	// Field 13 -- Info field which contains name/value pairs
	if td.lastLiveQueryTimeInt64 != nil {
		err = encoder.EncodeMap(map[string]interface{}{
			"LastLiveQueryTime": *td.lastLiveQueryTimeInt64,
		})
		if err != nil {
			return
		}
	}

	if td.lastLiveQueryTimeUint64 != nil {
		err = encoder.EncodeMap(map[string]interface{}{
			"LastLiveQueryTime": *td.lastLiveQueryTimeUint64,
		})
		if err != nil {
			return
		}
	}
	return hex.EncodeToString(buf.Bytes()), nil
}

func generateMsgPackParseTestFromTD1(rhd RequestHeaderDecoder, expectedVersion int, expectedLastUpdateTime, expectedLastLiveQueryTime *time.Time, expectedErr error) func() {
	return func() {
		msgpackHeaderHexString, _ := testData{version: expectedVersion, lastUpdateTime: timePtrToIntPtr(expectedLastUpdateTime), lastLiveQueryTimeInt64: timePtrToIntPtr(expectedLastLiveQueryTime), expectErr: expectedErr}.Encode()
		generateMsgPackParseTest(rhd, msgpackHeaderHexString, expectedVersion, expectedLastUpdateTime, expectedLastLiveQueryTime, expectedErr)
	}
}

func generateMsgPackParseTestFromTD2(rhd RequestHeaderDecoder, expectedVersion int, expectedLastUpdateTime, expectedLastLiveQueryTime *time.Time, expectedErr error) func() {
	return func() {
		msgpackHeaderHexString, _ := testData{version: expectedVersion, lastUpdateTime: timePtrToIntPtr(expectedLastUpdateTime), lastLiveQueryTimeUint64: timePtrToUIntPtr(expectedLastLiveQueryTime), expectErr: expectedErr}.Encode()
		generateMsgPackParseTest(rhd, msgpackHeaderHexString, expectedVersion, expectedLastUpdateTime, expectedLastLiveQueryTime, expectedErr)
	}
}

func generateMultipartMsgPack(expectedVersion int, expectedLastUpdateTime, expectedLastLiveQueryTime *time.Time) string {
	encodedHex, _ := testData{version: expectedVersion, lastUpdateTime: timePtrToIntPtr(expectedLastUpdateTime), lastLiveQueryTimeUint64: timePtrToUIntPtr(expectedLastLiveQueryTime), expectErr: nil}.Encode()
	rawBytes, _ := hex.DecodeString(encodedHex)
	boundary := "--hello-boundary"
	var buf bytes.Buffer

	// generate multipart
	buf.WriteString(fmt.Sprintf("%s\n\n", boundary))
	buf.WriteString(`Content-Disposition: form-data; name="other"\n\n`)
	buf.WriteString("junk text")
	buf.WriteString(fmt.Sprintf("%s\n", boundary))
	buf.WriteString(`Content-Disposition: form-data; name="message"\n\n`)
	buf.Write(rawBytes)
	buf.WriteString(fmt.Sprintf("%s--\n\n", boundary))
	return hex.EncodeToString(buf.Bytes())
}

func generateMsgPackParseTest(rhd RequestHeaderDecoder, msgPackHeaderHexString string, expectedVersion int, expectedLastUpdateTime, expectedLastLiveQueryTime *time.Time, expectedErr error) func() {
	return func() {
		msgpackHeaderBytes, err := hex.DecodeString(msgPackHeaderHexString)
		Expect(err).To(BeNil())

		version, lastUpdateTime, lastLiveQueryTime, err := rhd.ParseMsgPack(bytes.NewReader(msgpackHeaderBytes))
		logger.BaseLogger().Info(fmt.Sprintf("Parsed msg pack: %s", lastLiveQueryTime))

		if expectedErr == nil {
			Expect(err).To(BeNil())
		} else {
			didErrorsMatch := errors.Is(err, expectedErr)
			Expect(didErrorsMatch).To(BeTrue())
			return
		}
		Expect(version).To(Equal(expectedVersion))
		Expect(lastLiveQueryTime).To(Equal(expectedLastLiveQueryTime))
		Expect(lastUpdateTime).To(Equal(expectedLastUpdateTime))
	}
}

func TestMsgPackService(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "msgpack requestHeaderDecoder")
}

// ----- Begin Tests ----------------------------------------

var _ = Describe("MsgPack RequestHeaderDecoder", func() {
	rhd := NewRequestHeaderDecoder()

	lut := time.UnixMilli(1580644800000)  // Sunday, February 2, 2020 12:00:00 GMT
	llqt := time.UnixMilli(1568030400000) // Monday, September 9, 2019 12:00:00 GMT

	// NOTE: These tests will generate valid headers with specific data configurations
	Context("Datagen 'tests' (utilities)", func() {
		It("Full header with last update and livequery(int64)", func() {
			msgpackHeaderHexString, _ := testData{version: 17, lastUpdateTime: timePtrToIntPtr(&lut), lastLiveQueryTimeInt64: timePtrToIntPtr(&llqt)}.Encode()
			logger.BaseLogger().Info("Full Header (version, last update, livequery-int64)", zap.String("header", msgpackHeaderHexString))
		})

		It("Full header with last update and livequery(uint64)", func() {
			msgpackHeaderHexString, _ := testData{version: 17, lastUpdateTime: timePtrToIntPtr(&lut), lastLiveQueryTimeUint64: timePtrToUIntPtr(&llqt)}.Encode()
			logger.BaseLogger().Info("Full Header (version, last update, livequery-uint64)", zap.String("header", msgpackHeaderHexString))
		})

		It("Header with version & livequery(int64), no last update", func() {
			msgpackHeaderHexString, _ := testData{version: 17, lastLiveQueryTimeInt64: timePtrToIntPtr(&llqt)}.Encode()
			logger.BaseLogger().Info("Partial Header (version, livequery-int64)", zap.String("header", msgpackHeaderHexString))
		})

		It("Header with version & last update, no livequery", func() {
			msgpackHeaderHexString, _ := testData{version: 17, lastUpdateTime: timePtrToIntPtr(&lut)}.Encode()
			logger.BaseLogger().Info("Partial Header (version, last update)", zap.String("header", msgpackHeaderHexString))
		})

		It("Header with version only", func() {
			msgpackHeaderHexString, _ := testData{version: 17}.Encode()
			logger.BaseLogger().Info("Partial Header (version)", zap.String("header", msgpackHeaderHexString))
		})
	})

	Context("valid headers", func() {
		It("should parse live query(as int64) and last update time when version is greater than or equal to 17", generateMsgPackParseTestFromTD1(
			rhd, 17, &lut, &llqt, nil,
		))

		It("should parse live query(as uint64) and last update time when version is greater than or equal to 17", generateMsgPackParseTestFromTD2(
			rhd, 17, &lut, &llqt, nil,
		))

		It("should not error if no live query time is provided", generateMsgPackParseTestFromTD1(
			rhd,
			17,
			&lut,
			nil,
			nil,
		))
		It("should not parse live query time when version is less than 17", generateMsgPackParseTestFromTD1(
			rhd,
			16,
			&lut,
			nil,
			nil,
		))
	})

	Context("invalid header", func() {
		It("should reject empty headers", generateMsgPackParseTest(
			rhd,
			"",
			0,
			&lut,
			nil,
			io.EOF,
		))
		It("should reject headers without last update time", generateMsgPackParseTestFromTD1(
			rhd,
			17,
			nil,
			nil,
			nil,
		))
	})
	It("should generate a multipart body with msgpack", func() {
		body := generateMultipartMsgPack(17, &lut, &llqt)
		Expect(len(body)).To(BeComparableTo(420))
	})
})
