// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package deviceresponseheader

import (
	"io"
	"strings"
	"time"

	"github.com/vmihailenco/msgpack/v5"
)

const (
	// HelloResponseCode -- As per: https://gitlab.bit9.local/psc/platform/ems/hinting/-/blob/main/pkg/hints/messagecode/message_code.go
	HelloResponseCode = 0x201d
)

type (
	ResponseHeaderDecoder interface {
		Decode(hexEncodedResponse string) (responseHeader *DeviceResponseHeader, err error)
	}

	responseHeaderDecoder struct {
		builder msgPackDecoderBuilder
	}

	msgPackDecoder interface {
		DecodeInt64() (int64, error)
		Skip() error
		DecodeBool() (bool, error)
		DecodeString() (string, error)
		Decode(interface{}) error
		DecodeArrayLen() (int, error)
	}

	msgPackDecoderBuilder func(reader io.Reader) msgPackDecoder
)

func NewResponseHeaderDecoder() responseHeaderDecoder {
	return responseHeaderDecoder{
		builder: func(r io.Reader) msgPackDecoder {
			return msgpack.NewDecoder(r)
		},
	}
}

// Decode -- Decode a response msgpack object into a DeviceResponseHeader for testing/validation purposes (not used by the product itself)
//   - See https://confluence.carbonblack.local/display/ENDOC/Device+Request+Protocol#DeviceRequestProtocol-5.3ResponseHeader
//
// NOTE: Data is stored by position number, not field name!
func (responseHeaderDecoder *responseHeaderDecoder) Decode(msgPackEncodedString string) (responseHeader *DeviceResponseHeader, err error) {
	responseHeader = new(DeviceResponseHeader)
	decoder := msgpack.NewDecoder(strings.NewReader(msgPackEncodedString))
	if responseHeader.Version, err = decoder.DecodeInt64(); err != nil { // Field 1
		return nil, err
	}
	if responseHeader.Code, err = decoder.DecodeInt64(); err != nil { // Field 2
		return nil, err
	}
	err = decoder.Skip()
	if err != nil { // Field 3 (msgId) skipped since we skip it on write
		return nil, err
	}
	if responseHeader.Success, err = decoder.DecodeBool(); err != nil { // Field 4
		return nil, err
	}
	if !responseHeader.Success {
		if responseHeader.ErrorCode, err = decoder.DecodeInt64(); err != nil { // Field 5
			return nil, err
		}
	} else {
		err = decoder.Skip()
		if err != nil { // defaults to 0
			return nil, err
		}
	}
	if responseHeader.Message, err = decoder.DecodeString(); err != nil { // Field 6
		return nil, err
	}

	if err := decoder.Decode(&responseHeader.Hints); err != nil { // Field 7
		return nil, err
	}

	var lastUpdateTimeInt int64
	if lastUpdateTimeInt, err = decoder.DecodeInt64(); err != nil { // Field 8
		return
	}

	if lastUpdateTimeInt != 0 {
		lastUpdateTime := time.UnixMilli(lastUpdateTimeInt)
		responseHeader.LastUpdateTime = &lastUpdateTime
	}

	if responseHeader.Version >= 17 { // Field 9
		lastLiveQueryTimeInt, err := decoder.DecodeInt64()
		if err != nil {
			if err.Error() == "EOF" {
				// lastLiveQueryTime not found in msgPackEncodedString
				err = nil
			} else {
				return nil, err
			}
		}

		if lastLiveQueryTimeInt != 0 {
			lastLiveQueryTime := time.UnixMilli(lastLiveQueryTimeInt)
			responseHeader.LastLiveQueryTime = &lastLiveQueryTime
		}
	}
	return responseHeader, nil
}
