// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package deviceresponseheader

import (
	"fmt"
	"time"

	"hello-api/internal/logger"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

const (
	millis1 = 1672617600000 // Monday, January 2, 2023 0:00:00
	millis2 = 1657238400000 // Friday, July 8, 2022 0:00:00
)

var _ = Describe("msg pack response decoder", func() {
	decoder := NewResponseHeaderDecoder()
	It("Validates msgpack transparency (encode -> decode)", func() {
		s := NewResponseHeaderEncoder()
		lut := time.UnixMilli(millis1)
		llqt := time.UnixMilli(millis2)
		testData := NewDeviceResponseHeader(21, 33, true, 0, "Tested", []int{0, 2, 4}, &lut, &llqt)
		msgPackEncodedString, err := s.Encode(testData)
		Expect(err).To(BeNil())

		object, err := decoder.Decode(msgPackEncodedString)
		Expect(err).To(BeNil())
		Expect(object).ToNot(BeNil())

		Expect(object.Version).To(BeComparableTo(int64(21)))
		Expect(object.Code).To(BeComparableTo(int64(33)))
		Expect(object.Success).To(BeComparableTo(true))
		Expect(object.ErrorCode).To(BeComparableTo(int64(0)))
		Expect(object.Message).To(BeComparableTo("Tested"))
		Expect(len(object.Hints)).To(BeComparableTo(3))
		Expect(object.Hints[0]).To(BeComparableTo(0))
		Expect(object.Hints[1]).To(BeComparableTo(2))
		Expect(object.Hints[2]).To(BeComparableTo(4))
		Expect(object.LastUpdateTime.UnixMilli()).To(BeComparableTo(int64(millis1)))
		Expect(object.LastLiveQueryTime.UnixMilli()).To(BeComparableTo(int64(millis2)))
	})

	It("Validates msgpack transparency (encode -> decode), with error", func() {
		s := NewResponseHeaderEncoder()
		lut := time.UnixMilli(millis1)
		llqt := time.UnixMilli(millis2)
		testData := NewDeviceResponseHeader(21, 33, false, 42, "Tested", []int{0, 2, 4}, &lut, &llqt)
		msgPackEncodedString, err := s.Encode(testData)
		Expect(err).To(BeNil())

		object, err := decoder.Decode(msgPackEncodedString)
		Expect(err).To(BeNil())
		Expect(object).ToNot(BeNil())

		// all we care about for this test
		Expect(object.Success).To(BeComparableTo(false))
		Expect(object.ErrorCode).To(BeComparableTo(int64(42)))
	})

	It("Validates msgpack transparency (encode -> decode), if no error we ignore ErrorCode", func() {
		s := NewResponseHeaderEncoder()
		lut := time.UnixMilli(millis1)
		llqt := time.UnixMilli(millis2)
		testData := NewDeviceResponseHeader(21, 33, true, 42, "Tested", []int{0, 2, 4}, &lut, &llqt)
		msgPackEncodedString, err := s.Encode(testData)
		Expect(err).To(BeNil())

		object, err := decoder.Decode(msgPackEncodedString)
		Expect(err).To(BeNil())
		Expect(object).ToNot(BeNil())

		// all we care about for this test
		Expect(object.Success).To(BeComparableTo(true))
		Expect(object.ErrorCode).To(BeComparableTo(int64(0)))
	})

	It("Validates msgpack transparency (encode -> decode) empty hints is nil", func() {
		s := NewResponseHeaderEncoder()
		lut := time.UnixMilli(millis1)
		llqt := time.UnixMilli(millis2)
		testData := NewDeviceResponseHeader(21, 33, true, 42, "Tested", []int{}, &lut, &llqt)
		msgPackEncodedString, err := s.Encode(testData)
		Expect(err).To(BeNil())

		object, err := decoder.Decode(msgPackEncodedString)
		Expect(err).To(BeNil())
		Expect(object).ToNot(BeNil())
		Expect(object.Hints).ToNot(BeNil())

		// all we care about for this test
		Expect(len(object.Hints)).To(BeComparableTo(0))
	})

	It("rejects bad input", func() {
		obj, err := decoder.Decode("afdasfsda")
		Expect(err).ToNot(BeNil())
		// NOTE: Decode will parse until it is complete or has an error; while the  object might be returned on an error, it should not be trusted.
		logger.BaseLogger().Info(fmt.Sprintf("Object returned was: %+v", obj))
	})
})
