// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package deviceresponseheader

import (
	"bytes"
	"fmt"
	"io"
	"time"

	"github.com/vmihailenco/msgpack/v5"
	"go.uber.org/zap/zapcore"
)

// DeviceResponseHeader -- structure to hold a Hello response information prior to encoding into a msgpack for delivery to the calling device
// - See: https://confluence.carbonblack.local/display/ENDOC/Device+Request+Protocol#DeviceRequestProtocol-5.3ResponseHeader
// NOTE: fields made exported for ease of access; msgpack uses data position in the byte stream, not field name, for identification
type DeviceResponseHeader struct {
	Version           int64
	Code              int64
	Success           bool
	ErrorCode         int64
	Message           string
	Hints             []int
	LastUpdateTime    *time.Time
	LastLiveQueryTime *time.Time
}

func (drh *DeviceResponseHeader) MarshalLogObject(enc zapcore.ObjectEncoder) (err error) {
	enc.AddInt64("version", drh.Version)
	enc.AddInt64("code", drh.Code)
	enc.AddBool("success", drh.Success)
	enc.AddInt64("error_code", drh.ErrorCode)
	enc.AddString("message", drh.Message)
	_ = enc.AddArray("hints", zapcore.ArrayMarshalerFunc(func(enc zapcore.ArrayEncoder) error {
		for _, hint := range drh.Hints {
			enc.AppendInt(hint)
		}
		return nil
	}))

	if drh.LastUpdateTime != nil {
		enc.AddTime("last_update_time", *drh.LastUpdateTime)
	} else {
		enc.AddString("last_update_time", "null")
	}
	if drh.LastLiveQueryTime != nil {
		enc.AddTime("last_live_query_time", *drh.LastLiveQueryTime)
	} else {
		enc.AddString("last_live_query_time", "null")
	}
	return nil
}

func NewDeviceResponseHeader(version, code int64, success bool, errorCode int64, message string, hints []int, lastUpdateTime, lastLiveQueryTime *time.Time) *DeviceResponseHeader {
	return &DeviceResponseHeader{
		Version: version, Code: code, Success: success, ErrorCode: errorCode, Message: message,
		Hints: hints, LastUpdateTime: lastUpdateTime, LastLiveQueryTime: lastLiveQueryTime,
	}
}

type (
	// ResponseHeaderEncoder is the interface to parse msgpack headers
	ResponseHeaderEncoder interface {
		Encode(headerInfo *DeviceResponseHeader) (encodedResponse string, err error)
	}

	messagePackEncoder interface {
		Encode(v interface{}) error
		EncodeInt64(n int64) error
		EncodeNil() error
		EncodeBool(n bool) error
		EncodeString(n string) error
		EncodeArrayLen(n int) error
		EncodeMap(m map[string]interface{}) error
	}
	encoderBuilder        func(b io.Writer) messagePackEncoder
	responseHeaderEncoder struct{ builder encoderBuilder }
)

func NewResponseHeaderEncoder() ResponseHeaderEncoder {
	return &responseHeaderEncoder{builder: func(b io.Writer) messagePackEncoder { return msgpack.NewEncoder(b) }}
}

func NewResponseHeaderEncoderFromBuilder(builder encoderBuilder) ResponseHeaderEncoder {
	return &responseHeaderEncoder{builder: builder}
}

// Encode -- Encodes hello response information into a msgpack object
//   - See https://confluence.carbonblack.local/display/ENDOC/Device+Request+Protocol#DeviceRequestProtocol-5.3ResponseHeader
//
// NOTE: Data is stored by position number, not field name!
func (s *responseHeaderEncoder) Encode(headerInfo *DeviceResponseHeader) (msgPackEncodedString string, err error) {
	var buffer bytes.Buffer
	encoder := s.builder(&buffer)
	if headerInfo == nil {
		err = fmt.Errorf("no device response header was provided")
		return
	}

	// Field 1 -- version of the response
	if err = encoder.EncodeInt64(headerInfo.Version); err != nil {
		err = fmt.Errorf("failed to encode version: %w", err)
		return
	}

	// Field 2 -- Code identifying the message type
	if err = encoder.EncodeInt64(headerInfo.Code); err != nil {
		err = fmt.Errorf("failed to encode message code: %w", err)
		return
	}

	// Field 3 -- Unique message identifier optionally set by the client side (ignored)
	if err = encoder.EncodeNil(); err != nil {
		err = fmt.Errorf("failed to encode nil message identifier: %w", err)
		return
	}

	// Field 4 -- Boolean indicator whether the request succeeded or not
	if err = encoder.EncodeBool(headerInfo.Success); err != nil {
		err = fmt.Errorf("failed to encode request success: %w", err)
		return
	}

	// Field 5 -- Error code; must be set
	if err = encoder.EncodeInt64(headerInfo.ErrorCode); err != nil {
		err = fmt.Errorf("failed to encode error code: %w", err)
		return
	}

	// Field 6 -- Informational message to describe the success or failure of the request
	if err = encoder.EncodeString(headerInfo.Message); err != nil {
		err = fmt.Errorf("failed to encode request informational message: %w", err)
		return
	}

	// Field 7 -- Array of hints that informs the agent of information the agent should request from the cloud
	if headerInfo.Hints != nil {
		if err = encoder.Encode(headerInfo.Hints); err != nil {
			err = fmt.Errorf("failed to encode hints array: %w", err)
			return
		}
	} else {
		if err = encoder.EncodeNil(); err != nil {
			err = fmt.Errorf("failed to encode nil hints: %w", err)
			return
		}
	}

	// Field 8 -- Timestamp of when the last time the sensor received a hint indicator to pull something new from the cloud
	if headerInfo.LastUpdateTime != nil {
		if err = encoder.EncodeInt64(headerInfo.LastUpdateTime.UnixMilli()); err != nil {
			err = fmt.Errorf("failed to encode last update time: %w", err)
			return
		}
	} else {
		if err = encoder.EncodeNil(); err != nil {
			err = fmt.Errorf("failed to encode nil last update time: %w", err)
			return
		}
	}

	// Field 9 -- Timestamp of when the last time the sensor received a hint indicator to pull pending live queries from the cloud
	if headerInfo.LastLiveQueryTime != nil {
		if err = encoder.EncodeInt64(headerInfo.LastLiveQueryTime.UnixMilli()); err != nil {
			err = fmt.Errorf("failed to encode last live query time: %w", err)
			return
		}
	} else {
		if err = encoder.EncodeNil(); err != nil {
			err = fmt.Errorf("failed to encode nil last live query time: %w", err)
			return
		}
	}
	return buffer.String(), nil
}
