// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package deviceresponseheader

import (
	"io"
	"testing"
	"time"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"github.com/vmihailenco/msgpack/v5"
)

type spyCall struct {
	MethodName string
	Arg        any
	Retval     any
}

type spyEncoder struct {
	realEncoder messagePackEncoder
	calls       []spyCall
}

type spyEncoderBuilder struct {
	currentSpy *spyEncoder
}

func (spyEncoderBuilder *spyEncoderBuilder) build(w io.Writer) messagePackEncoder {
	realEncoder := msgpack.NewEncoder(w)
	spy := &spyEncoder{calls: make([]spyCall, 0), realEncoder: realEncoder}
	spyEncoderBuilder.currentSpy = spy
	return spy
}

func (s *spyEncoder) addSpiedCall(methodName string, arg, returnValue any) {
	s.calls = append(s.calls, spyCall{MethodName: methodName, Arg: arg, Retval: returnValue})
}

func (s *spyEncoder) Encode(v interface{}) error {
	retValue := s.realEncoder.Encode(v)
	s.addSpiedCall("Encode", v, retValue)
	return retValue
}

func (s *spyEncoder) EncodeInt64(n int64) error {
	retValue := s.realEncoder.EncodeInt64(n)
	s.addSpiedCall("EncodeInt64", n, retValue)
	return retValue
}

func (s *spyEncoder) EncodeNil() error {
	retValue := s.realEncoder.EncodeNil()
	s.addSpiedCall("EncodeNil", nil, retValue)
	return retValue
}

func (s *spyEncoder) EncodeBool(n bool) error {
	retValue := s.realEncoder.EncodeBool(n)
	s.addSpiedCall("EncodeBool", n, retValue)
	return retValue
}

func (s *spyEncoder) EncodeString(n string) error {
	retValue := s.realEncoder.EncodeString(n)
	s.addSpiedCall("EncodeString", n, retValue)
	return retValue
}

func (s *spyEncoder) EncodeArrayLen(n int) error {
	retValue := s.realEncoder.EncodeArrayLen(n)
	s.addSpiedCall("EncodeArrayLen", n, retValue)
	return retValue
}

func (s *spyEncoder) EncodeMap(m map[string]interface{}) error {
	retValue := s.realEncoder.EncodeMap(m)
	s.addSpiedCall("EncodeMap", m, retValue)
	return retValue
}

var _ = Describe("ResponseHeader Encoder", func() {
	s := NewResponseHeaderEncoder()
	now := time.Now()
	time1 := time.UnixMilli(1672617600000) // Monday, January 2, 2023 0:00:00
	time2 := time.UnixMilli(1657238400000) // Friday, July 8, 2022 0:00:00

	Context("Encodes valid inputs", func() {
		// NOTE: For encoding tests, no simple way to validate since we have no decoder

		It("Accepts input as from the hello controller", func() {
			testData := NewDeviceResponseHeader(21, 0x021, true, 0, "", []int{}, nil, nil)
			encodedString, err := s.Encode(testData)
			Expect(err).To(BeNil())
			Expect(encodedString).ToNot(BeNil())
		})

		It("Accepts fully valid input", func() {
			test := &DeviceResponseHeader{
				Version:           789,
				Code:              456,
				Success:           false,
				ErrorCode:         123,
				Message:           "Hello world",
				Hints:             []int{0, 2, 3, 4},
				LastUpdateTime:    &now,
				LastLiveQueryTime: &now,
			}
			encodedString, err := s.Encode(test)
			Expect(err).To(BeNil())
			Expect(encodedString).ToNot(BeNil())
		})

		It("Accepts minimal fully valid input", func() {
			test := DeviceResponseHeader{
				Version:           0,
				Code:              0,
				Success:           true,
				ErrorCode:         0,
				Message:           "",
				Hints:             []int{},
				LastUpdateTime:    nil,
				LastLiveQueryTime: nil,
			}
			encodedString, err := s.Encode(&test)
			Expect(err).To(BeNil())
			Expect(encodedString).ToNot(BeNil())
		})

		It("calls encode methods in appropriate order", func() {
			spyBuilder := &spyEncoderBuilder{}
			spyService := NewResponseHeaderEncoderFromBuilder(spyBuilder.build)
			test := DeviceResponseHeader{
				Version:           2023,
				Code:              1337,
				Success:           true,
				ErrorCode:         0,
				Message:           "MESSAGE",
				Hints:             []int{0, 2, 3},
				LastUpdateTime:    &time1,
				LastLiveQueryTime: &time2,
			}
			expectedCalls := []spyCall{
				{"EncodeInt64", test.Version, nil},
				{MethodName: "EncodeInt64", Arg: test.Code, Retval: nil},
				{MethodName: "EncodeNil", Arg: nil, Retval: nil},
				{MethodName: "EncodeBool", Arg: test.Success, Retval: nil},
				{MethodName: "EncodeInt64", Arg: int64(0), Retval: nil},
				{MethodName: "EncodeString", Arg: test.Message, Retval: nil},
				{MethodName: "Encode", Arg: test.Hints, Retval: nil},
				{MethodName: "EncodeInt64", Arg: test.LastUpdateTime.UnixMilli(), Retval: nil},
				{MethodName: "EncodeInt64", Arg: test.LastLiveQueryTime.UnixMilli(), Retval: nil},
			}
			encodedString, err := spyService.Encode(&test)
			currentSpy := spyBuilder.currentSpy
			actualCalls := currentSpy.calls
			Expect(err).To(BeNil())
			Expect(encodedString).ToNot(BeNil())
			Expect(actualCalls).To(BeComparableTo(expectedCalls))
		})

		It("accepts nil hints input", func() {
			test := DeviceResponseHeader{
				Version:           18,
				Code:              8221,
				Success:           false,
				ErrorCode:         401,
				Message:           "Invalid request",
				Hints:             nil,
				LastUpdateTime:    nil,
				LastLiveQueryTime: nil,
			}
			encodedString, err := s.Encode(&test)
			Expect(encodedString).ToNot(BeNil())
			Expect(err).To(BeNil())
			decoder := NewResponseHeaderDecoder()
			object, err := decoder.Decode(encodedString)
			Expect(err).NotTo(HaveOccurred())
			Expect(*object).To(BeEquivalentTo(test))
		})
	})

	// NOTE: Unable to test negative conditions of encoding since the go structure prevents this.
})

func TestMsgPackService(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "msgpack  response encoder+decoder test")
}
