// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package app

import (
	"context"
	"errors"
	"fmt"
	"net"
	"net/http"
	"sync/atomic"
	"time"

	"hello-api/internal/app/controllers"
	"hello-api/internal/app/middleware"
	"hello-api/internal/logger"

	"github.com/gin-gonic/gin"
	"go.uber.org/zap"
)

func SetupRouter(service *HelloService) *gin.Engine {
	logger.BaseLogger().Debug("Setting up router...")

	router := gin.New()
	mwf := middleware.NewMiddlewareFactory(service.Logger, service.metricsClient, service.cfg)

	router.Group("/hello-api").GET("healthCheck", func(ctx *gin.Context) {
		isReadyCheck := ctx.Param("isReady")
		if isReadyCheck == "true" {
			if atomic.LoadInt32(service.Ready) == 1 {
				ctx.Status(http.StatusOK)
				return
			} else {
				ctx.Status(http.StatusServiceUnavailable)
				return
			}
		}
		if atomic.LoadInt32(service.alive) == 1 {
			ctx.Status(http.StatusOK)
			return
		}
		ctx.Status(http.StatusServiceUnavailable)
	})

	var promMetrics *middleware.Metrics
	if service.cfg.PrometheusEnabled {
		promMetrics = middleware.RegisterPromMetrics()
	}

	helloController := controllers.NewHelloController(service.cfg, service.metricsClient)
	servicesPath := router.Group("/services").Use(mwf.GetTracingChain()...).Use(middleware.REDMetrics(promMetrics))
	helloController.AddToRouter(servicesPath, middleware.LctPublisherMiddleware(service.lctProducer, service.metricsClient), middleware.HintingMiddleware(service.hintsClient, service.metricsClient))
	logger.BaseLogger().Debug("Finished setting up router...")
	return router
}

type HttpServer struct {
	ginEngine  *gin.Engine
	httpServer *http.Server
	Port       string
}

func NewHttpServer(service *HelloService) (*HttpServer, error) {
	router := SetupRouter(service)

	if service.cfg.ProfilingEnabled {
		middleware.RouteProfiler(router, service.cfg)
	}

	httpServer := &http.Server{
		Addr:              service.cfg.ServerAddr,
		Handler:           router,
		ReadHeaderTimeout: service.cfg.HttpReadTimeout,
		ReadTimeout:       service.cfg.HttpReadTimeout,
		WriteTimeout:      service.cfg.HttpWriteTimeout,
	}
	listener, err := net.Listen("tcp", service.cfg.ServerAddr)
	if err != nil {
		return nil, err
	}
	go func() {
		if err := httpServer.Serve(listener); err != nil && !errors.Is(err, http.ErrServerClosed) {
			logger.BaseLogger().Error("listen: %s\n", zap.Error(err))
		}
	}()
	return &HttpServer{ginEngine: router, httpServer: httpServer, Port: fmt.Sprintf("%d", listener.Addr().(*net.TCPAddr).Port)}, nil
}

func (server *HttpServer) Shutdown() error {
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	if err := server.httpServer.Shutdown(ctx); err != nil {
		return err
	}
	return nil
}
