// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package config

import (
	"flag"
	"fmt"
	"os"
	"sync"
	"time"

	"github.com/spf13/viper"
)

// singleton configuration instance
var (
	onceConfig   sync.Once
	instance     *Config
	profilerFlag *bool
)

func ParseCliFlags() {
	profilerFlag = flag.Bool("profiler", false, "Enable profiling")
	flag.Parse()
}

// Actual method to load the configuration from the local environment, over any existing defaults.
func reset() {
	cfg, err := newConfigFromEnv(true)
	if err != nil {
		panic(fmt.Sprintf("Failed to load config from environment: %v", err))
	}
	instance = cfg
}

// GetInstance -- Get the actual configuration
func GetInstance() *Config {
	onceConfig.Do(func() {
		reset()
	})
	return instance
}

// ===== Configuration Constants ==================================================
//

// Environment Variable Names
const (
	envarKeyServerAddr = "SERVER_ADDR"

	// METRICS
	envVarKeyWavefrontEnabled  = "WAVEFRONT_ENABLED"
	envVarKeyWavefrontEndpoint = "WAVEFRONT_ENDPOINT"
	envWavefrontPort           = "WAVEFRONT_PORT"
	envWavefrontTracePort      = "WAVEFRONT_TRACE_PORT"
	envVarKeyMetricInterval    = "METRIC_INTERVAL_SECONDS"
	envVarKeyMetricSamplerRate = "METRIC_SAMPLER_RATE"

	envVarKeyPrometheusEnabled = "PROMETHEUS_ENABLED"
	envVarKeyPrometheusPort    = "PROMETHEUS_PORT"

	// MISC
	envVarEnvCategory      = "ENV_CATEGORY"
	envVarKeyCommitSHA     = "COMMIT_SHA"
	envVarKeyLogLevel      = "LOG_LEVEL"
	envVarKeyServicePrefix = "SERVICE_PREFIX"

	// HTTP
	envVarKeyReadTimeout   = "HTTP_READ_TIMEOUT_MS"
	envVarKeyHandleTimeout = "HTTP_HANDLE_TIMEOUT_MS"
	envVarKeyWriteTimeout  = "HTTP_WRITE_TIMEOUT_MS"

	// LCT
	envVarKeyLCTStreamName = "LCT_STREAM_NAME"
)

// Wavefront Constants
const (
	baseDimProduct     = "psc"
	baseDimApplication = "ems-device-hello"
	baseDimService     = "ems-device-hello"
)

// Configuration defaults
const (
	// metrics
	defaultMetricInterval    = 10
	defaultMetricSamplerRate = 5

	// Misc.
	defaultEnvCategory   = "local"
	defaultLogLevel      = "info"
	defaultServicePrefix = "hello"

	// http
	defaultReadTimeout   = 5000
	defaultHandleTimeout = 5000
	defaultWriteTimeout  = 5000
	defaultServerAddr    = "0.0.0.0:5004"

	// lct
	defaultStreamName = "local-ems-device-hello-local-infra-lct-stream-psc-ems-last-contact-time-updater-stream"
)

//
// ================================================== Configuration Constants =====

type IConfigParam[T any] interface {
	bind()
	establishDefault()
	Get(bind bool) T
}

type GenericEnvConfigParam[T any] struct {
	envar        string
	defaultValue T
	hasDefault   bool
}
type StringParam struct {
	GenericEnvConfigParam[string]
}

func (sp StringParam) Get(bind bool) string {
	if bind {
		sp.bind()
	}
	return viper.GetString(sp.envar)
}

type DurationParam struct {
	GenericEnvConfigParam[time.Duration]
	unit time.Duration
}

func (dp DurationParam) Get(bind bool) time.Duration {
	if bind {
		dp.bind()
	}
	return viper.GetDuration(dp.envar) * dp.unit
}

func NewConfigParam(envar string) IConfigParam[string] {
	return StringParam{GenericEnvConfigParam: NewGenericConfigParamWithoutDefault(envar)}
}

func NewConfigParamWithDefault(envar, defaultValue string) IConfigParam[string] {
	return StringParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

func NewDurationConfigParam(envar string, defaultValue, unit time.Duration) IConfigParam[time.Duration] {
	return DurationParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue), unit: unit}
}

type IntParam struct {
	GenericEnvConfigParam[int]
}

func (ip IntParam) Get(bind bool) int {
	if bind {
		ip.bind()
	}
	return viper.GetInt(ip.envar)
}

func NewIntParam(envar string, defaultValue int) IConfigParam[int] {
	return IntParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

type UIntParam struct {
	GenericEnvConfigParam[uint64]
}

func (ip UIntParam) Get(bind bool) uint64 {
	if bind {
		ip.bind()
	}
	return viper.GetUint64(ip.envar)
}

func NewUIntParam(envar string, defaultValue uint64) IConfigParam[uint64] {
	return UIntParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

type Int64Param struct {
	GenericEnvConfigParam[int64]
}

func (ip Int64Param) Get(bind bool) int64 {
	if bind {
		ip.bind()
	}
	return viper.GetInt64(ip.envar)
}

type BoolParam struct {
	GenericEnvConfigParam[bool]
}

func (ip BoolParam) Get(bind bool) bool {
	if bind {
		ip.bind()
	}
	return viper.GetBool(ip.envar)
}

func NewBoolParam(envar string, defaultValue bool) BoolParam {
	param := NewGenericConfigParam(envar, defaultValue)
	return BoolParam{param}
}

func (cp GenericEnvConfigParam[T]) bind() {
	_ = viper.BindEnv(cp.envar)
}

func (cp GenericEnvConfigParam[T]) establishDefault() {
	if cp.hasDefault {
		viper.SetDefault(cp.envar, cp.defaultValue)
	}
}

func NewGenericConfigParamWithoutDefault(envar string) GenericEnvConfigParam[string] {
	param := GenericEnvConfigParam[string]{envar: envar, hasDefault: false}
	return param
}

func NewGenericConfigParam[T any](envar string, defaultValue T) GenericEnvConfigParam[T] {
	param := GenericEnvConfigParam[T]{envar: envar, defaultValue: defaultValue, hasDefault: true}
	param.establishDefault()
	return param
}

type Config struct {
	ServerAddr string

	// METRICS
	WavefrontEnabled             bool
	WavefrontEndpoint            string
	WavefrontToken               string
	WavefrontPort                int
	WavefrontTracePort           int
	WavefrontHistogramPort       int
	WavefrontHistogramMinutePort int
	WavefrontHistogramHourPort   int
	WavefrontHistogramDayPort    int

	PrometheusEnabled bool
	PrometheusPort    int

	MetricPrefix      string
	MetricInterval    time.Duration
	MetricSamplerRate uint64
	MetricsEnabled    bool
	BaseDimensions    map[string]string

	// MISC
	CbdEnvName       string
	Region           string
	ServicePrefix    string
	ProfilingEnabled bool
	EnvCategory      string

	// HTTP
	HttpReadTimeout   time.Duration
	HttpWriteTimeout  time.Duration
	HttpHandleTimeout time.Duration

	// Hinting
	HintingUrl     string
	HintingTimeout time.Duration
	// HTTP HEADERS
	RequestIDHeader string

	// LCT STREAM NAME
	AwsEndpoint   string
	LctStreamName string

	// User Agent string for dev relations
	UserAgent string
	// Audit Mode for Dereg
	DeregAuditModeDisabled bool
}

func (cfg *Config) IsLocalEnv() bool {
	return cfg.EnvCategory == "local"
}

func (cfg *Config) IsDevEnv() bool {
	return cfg.EnvCategory == "dev"
}

func newConfigFromEnv(loadFromEnvironment bool) (*Config, error) {
	profilerEnabled := false
	if profilerFlag != nil {
		profilerEnabled = *profilerFlag
	}

	cfg := Config{
		ServerAddr: NewConfigParamWithDefault(envarKeyServerAddr, defaultServerAddr).Get(loadFromEnvironment),

		// METRICS
		WavefrontEnabled:   NewBoolParam(envVarKeyWavefrontEnabled, true).Get(loadFromEnvironment),
		WavefrontEndpoint:  NewConfigParam(envVarKeyWavefrontEndpoint).Get(loadFromEnvironment),
		WavefrontPort:      NewIntParam(envWavefrontPort, 2878).Get(loadFromEnvironment),
		WavefrontTracePort: NewIntParam(envWavefrontTracePort, 30000).Get(loadFromEnvironment),

		MetricPrefix:      "psc.ems-device-hello",
		MetricInterval:    NewDurationConfigParam(envVarKeyMetricInterval, defaultMetricInterval, time.Second).Get(loadFromEnvironment),
		MetricSamplerRate: NewUIntParam(envVarKeyMetricSamplerRate, defaultMetricSamplerRate).Get(loadFromEnvironment),

		PrometheusEnabled: NewBoolParam(envVarKeyPrometheusEnabled, true).Get(loadFromEnvironment),
		PrometheusPort:    NewIntParam(envVarKeyPrometheusPort, 2112).Get(loadFromEnvironment),

		// MISC
		CbdEnvName: NewConfigParamWithDefault("CBDENVNAME", "local").Get(loadFromEnvironment),

		ServicePrefix:    NewConfigParamWithDefault(envVarKeyServicePrefix, defaultServicePrefix).Get(loadFromEnvironment),
		ProfilingEnabled: profilerEnabled,
		EnvCategory:      NewConfigParamWithDefault(envVarEnvCategory, defaultEnvCategory).Get(loadFromEnvironment),

		// HTTP
		HttpReadTimeout:   NewDurationConfigParam(envVarKeyReadTimeout, defaultReadTimeout, time.Millisecond).Get(loadFromEnvironment),
		HttpWriteTimeout:  NewDurationConfigParam(envVarKeyWriteTimeout, defaultWriteTimeout, time.Millisecond).Get(loadFromEnvironment),
		HttpHandleTimeout: NewDurationConfigParam(envVarKeyHandleTimeout, defaultHandleTimeout, time.Millisecond).Get(loadFromEnvironment),

		// HTTP HEADERS
		RequestIDHeader: "Wf-Ot-Traceid",

		// LCT
		LctStreamName: NewConfigParamWithDefault(envVarKeyLCTStreamName, defaultStreamName).Get(loadFromEnvironment),
		AwsEndpoint:   NewConfigParam("AWS_ENDPOINT").Get(loadFromEnvironment),
		Region:        NewConfigParamWithDefault("REGION", "us-east-1").Get(loadFromEnvironment),

		// Hinting
		HintingUrl:     NewConfigParam("HINTING_URL").Get(loadFromEnvironment),
		HintingTimeout: NewDurationConfigParam("HINTING_TIMEOUT_MSECS", 500, time.Millisecond).Get(loadFromEnvironment),

		// Other
		UserAgent: "ems-device-hello/1.0.0",

		DeregAuditModeDisabled: NewBoolParam("DEREG_AUDIT_MODE_DISABLED", false).Get(loadFromEnvironment),
	}

	// NOTE: retained in case it is desired to add this at a later point.
	// sha := NewConfigParamWithDefault(envVarKeyCommitSHA, "").Get(loadFromEnvironment)
	// if sha != "" {
	// 	 cfg.UserAgent += fmt.Sprintf(" (%s)", sha)
	// }

	hostName, err := os.Hostname()
	if err != nil {
		return nil, err
	}
	cfg.BaseDimensions = map[string]string{
		"product":     baseDimProduct,
		"application": baseDimApplication,
		"service":     baseDimService,
		"env":         cfg.CbdEnvName,
		"instance":    hostName,
	}
	cfg.MetricsEnabled = cfg.WavefrontEndpoint != ""
	return &cfg, err
}
