// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hinting

import (
	"bytes"
	"encoding/json"
	"errors"
	"io"
	"net/http"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"hello-api/internal/config"
	"hello-api/internal/logger"
)

type (
	// Hinter -- Our simple Hinter interface for getting hints for a device
	Hinter interface {
		GetHints(hintRequest *hints.JsonHintRequest) (*hints.HintResult, error)
	}

	hintsClient struct {
		httpDoer
		url string
	}

	httpDoer func(req *http.Request) (*http.Response, error)
)

func NewHintsClient(client httpDoer, url string) Hinter {
	return &hintsClient{httpDoer: client, url: url}
}

// NewHintsClientFromConfig to be called from hinting middleware.
func NewHintsClientFromConfig(cfg *config.Config) Hinter {
	t := http.DefaultTransport.(*http.Transport).Clone()
	t.MaxIdleConns = 0
	t.MaxIdleConnsPerHost = 100
	client := http.Client{Timeout: cfg.HintingTimeout, Transport: t}
	return NewHintsClient(client.Do, cfg.HintingUrl)
}

func (hc *hintsClient) GetHints(jsonHintRequest *hints.JsonHintRequest) (hintResult *hints.HintResult, err error) {
	hintResult = new(hints.HintResult)
	if jsonHintRequest == nil {
		return hintResult, errors.New("no input provided")
	}
	var data []byte

	if data, err = json.Marshal(jsonHintRequest); err != nil {
		return
	}
	var req *http.Request
	if req, err = http.NewRequest("POST", hc.url, bytes.NewBuffer(data)); err != nil {
		return
	}
	req.Header.Set("Content-Type", "application/json")
	cfg := config.GetInstance()
	req.Header.Set("User-Agent", cfg.UserAgent)

	var response *http.Response
	response, err = hc.httpDoer(req)
	if err != nil {
		return
	}
	defer response.Body.Close()
	if response.StatusCode != http.StatusOK {
		return hintResult, errors.New("status code was not OK ")
	}
	var responseBytes []byte
	responseBytes, err = io.ReadAll(response.Body)
	if err != nil {
		return
	}
	err = json.Unmarshal(responseBytes, hintResult)
	if err != nil {
		logger.BaseLogger().Error(err.Error())
	}

	return
}
