// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hinting

import (
	"bytes"
	"errors"
	"io"
	"net/http"
	"testing"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

var t *testing.T

func TestHintingClient(test *testing.T) {
	RegisterFailHandler(Fail)
	t = test
	RunSpecs(test, "Hinting client tests")
}

const (
	orgkey = "7ZP4H5XR" // env = "local", orgId = 2023
)

var _ = Describe("Hinting client tests", func() {
	hr := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
	lastUpdateTime := time.UnixMilli(1675397106000)    // Friday, February 3, 2023 4:05:06
	lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
	hr.LastUpdateTime = &lastUpdateTime
	hr.LastLiveQueryTime = &lastLiveQueryTime
	hrData, _ := hr.MarshalJSON()

	version := 17
	requestLastUpdateTime := int64(0)
	requestLastLiveQueryTime := int64(0)
	messageCode := "REQ_HELLO"

	jsonHintRequest := &hints.JsonHintRequest{
		LastUpdateTime: &requestLastUpdateTime, LastLiveQueryTime: &requestLastLiveQueryTime, OrgKey: orgkey, DeviceGroupId: 456, DeviceId: 123, DeviceType: "WINDOWS", DeploymentType: "ENDPOINT", Version: &version, MessageCode: &messageCode,
	}

	// ----- Begin Tests --------------------------------------------------

	It("successfully retrieves hints", func() {
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBuffer(hrData))}, nil
		}, "ignored")

		hintResult, err := theClient.GetHints(jsonHintRequest)

		Expect(err).To(BeNil())
		Expect(hintResult).ToNot(BeNil())
		Expect(hintResult.Hints).To(BeEquivalentTo(hr.Hints))
		Expect(hintResult.LastUpdateTime).To(BeEquivalentTo(&lastUpdateTime))
		Expect(hintResult.LastLiveQueryTime).To(BeEquivalentTo(&lastLiveQueryTime))
	})

	It("returns an error when no hint request supplied", func() {
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBuffer(hrData))}, nil
		}, "ignored")

		hintResult, err := theClient.GetHints(nil)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	It("returns an error when hint response has an error", func() {
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBuffer(hrData))}, errors.New("bad call")
		}, "ignored")

		hintResult, err := theClient.GetHints(jsonHintRequest)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	It("returns an error when hint response has a 500 code", func() {
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 500, Body: io.NopCloser(bytes.NewBuffer(hrData))}, nil
		}, "ignored")

		hintResult, err := theClient.GetHints(jsonHintRequest)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	// FUTURE: Figure out how to make a test that has an error when reading the response body

	It("will return an error if the body can't be unmarshaled", func() {
		_, err := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBufferString("BOGUS"))}, nil
		}, "ignored").GetHints(jsonHintRequest)
		Expect(err).ToNot(BeNil())
	})
})
