// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package lct

import (
	"fmt"

	"go.uber.org/zap"
	"hello-api/internal/logger"
	"hello-api/internal/metrics"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	producer "github.com/zacharyestep/kinesis-producer"
	"google.golang.org/protobuf/proto"
)

type (

	// LctProducer -- Our simple LCT producer interface for publishing messages
	LctProducer interface {
		Publish(lct *devicelastcontacttime.DeviceLastContactTime) error
	}

	// StartStopableLctProducer -- We need extra methods at the service layer to start/stop and checking if the desired stream actually exists...
	StartStopableLctProducer interface {
		Start()
		Stop()
		LctProducer
	}

	FailureHandler func(<-chan error)

	// LctProducer implementation
	lctProducer struct {
		underlyingProducer underlyingProducerIface
		streamName         string
		FailureHandler
	}

	underlyingProducerIface interface {
		Put(data []byte, partitionKey string) error
		NotifyFailures() <-chan error
		Start()
		Stop()
	}

	kplLoggerAdapter struct {
		ourLogger logger.Logger
	}
)

func NewLctProducerFromConfig(producerConfig *producer.Config, metricsClient metrics.Service) (StartStopableLctProducer, error) {
	return NewLctProducer(producerConfig, defaultFailureHandler(metricsClient),
		func(config *producer.Config) (underlyingProducerIface, error) {
			return producer.New(config)
		})
}

func NewLctProducer(producerConfig *producer.Config, failureHandler FailureHandler, producerBuilder func(producerConfig *producer.Config) (underlyingProducerIface, error)) (StartStopableLctProducer, error) {
	kinesisProducer, err := producerBuilder(producerConfig)
	return &lctProducer{underlyingProducer: kinesisProducer, streamName: producerConfig.StreamName, FailureHandler: failureHandler}, err
}

func defaultFailureHandler(metricsClient metrics.Service) FailureHandler {
	return func(records <-chan error) {
		go func() {
			for r := range records {
				logger.BaseLogger().Error("Error while posting to kinesis", zap.Any("failure_record", r))
				if metricsClient != nil {
					metricsClient.Count("kpl.hello-lct-stream.failed", nil, 1)
				}
			}
		}()
	}
}

func (lctPublisher *lctProducer) Start() {
	lctPublisher.FailureHandler(lctPublisher.underlyingProducer.NotifyFailures())
	lctPublisher.underlyingProducer.Start()
}

func (lctPublisher *lctProducer) Stop() {
	lctPublisher.underlyingProducer.Stop()
}

func (lctPublisher *lctProducer) Publish(lct *devicelastcontacttime.DeviceLastContactTime) error {
	bytes, err := FormatLCTMessage(lct)
	if err != nil {
		return err
	}
	return lctPublisher.underlyingProducer.Put(bytes, fmt.Sprintf("%d", lct.GetDeviceId()))
}

func FormatLCTMessage(lct *devicelastcontacttime.DeviceLastContactTime) ([]byte, error) {
	return proto.Marshal(lct)
}

func NewProducerLogger(log logger.Logger) producer.Logger {
	return &kplLoggerAdapter{ourLogger: log}
}

func (pl *kplLoggerAdapter) Info(msg string, values ...producer.LogValue) {
	// supress info level logging to debug level to avoid log spam
	pl.ourLogger.Debug(msg, logValueToZapField(values...)...)
}

func (pl *kplLoggerAdapter) Error(msg string, err error, values ...producer.LogValue) {
	pl.ourLogger.Error(msg, append(logValueToZapField(values...), zap.Error(err))...)
}

func logValueToZapField(values ...producer.LogValue) []zap.Field {
	numValues := len(values)
	outputFields := make([]zap.Field, numValues)
	for i, inputValue := range values {
		outputFields[i] = zap.Any(inputValue.Name, inputValue.Value)
	}
	return outputFields
}
