// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package lct

import (
	"errors"
	"sync"
	"testing"

	"hello-api/internal/metrics"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
	"google.golang.org/protobuf/proto"

	producer "github.com/zacharyestep/kinesis-producer"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

func TestLctPublisher(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "LCT Publisher test")
}

const testStreamName = "someStreamName"

type spyMetricsService struct {
	counter   int64
	countName string
	wg        sync.WaitGroup
	metrics.Service
}

func (metricsService *spyMetricsService) Count(name string, _ map[string]string, value int64) {
	metricsService.counter += value
	metricsService.countName = name
	metricsService.wg.Done()
}

type mockUnderlyingProducer struct {
	putError          error
	putCalled         bool
	putCalledWithData []byte
	putCalledWithKey  string
	startCalled       bool
	stopCalled        bool
	failureChannel    chan error
}

func (mockProducer *mockUnderlyingProducer) Start() {
	mockProducer.startCalled = true
}

func (mockProducer *mockUnderlyingProducer) Stop() {
	mockProducer.stopCalled = true
	close(mockProducer.failureChannel)
}

func (mockProducer *mockUnderlyingProducer) Put(data []byte, partitionKey string) error {
	mockProducer.putCalled = true
	mockProducer.putCalledWithData = data
	mockProducer.putCalledWithKey = partitionKey
	if mockProducer.putError != nil {
		mockProducer.failureChannel <- nil
	}
	return mockProducer.putError
}

func (mockProducer *mockUnderlyingProducer) NotifyFailures() <-chan error {
	return mockProducer.failureChannel
}

var _ = Describe("lct publisher tests", Ordered, func() {
	mockUnderlyingPublisher := &mockUnderlyingProducer{failureChannel: make(chan error)}

	metricsService := spyMetricsService{}

	publisher, err := NewLctProducer(&producer.Config{StreamName: testStreamName}, defaultFailureHandler(&metricsService), func(cfg *producer.Config) (underlyingProducerIface, error) {
		return mockUnderlyingPublisher, nil
	})
	Expect(err).To(BeNil())
	publisher.Start()

	It("publishes records", func() {
		lct := &devicelastcontacttime.DeviceLastContactTime{LastContactTimestamp: 1337, DeviceId: 31337, OrgKey: "ORGKEY", DeregisterAfterTime: 30}
		marshalledLct, _ := proto.Marshal(lct)
		Expect(publisher.Publish(lct)).To(BeNil())
		Expect(mockUnderlyingPublisher.putCalled).To(BeComparableTo(true))
		Expect(mockUnderlyingPublisher.putCalledWithData).To(BeComparableTo(marshalledLct))
		Expect(mockUnderlyingPublisher.putCalledWithKey).To(BeComparableTo("31337"))
		Expect(metricsService.counter).To(BeIdenticalTo(int64(0)))
		Expect(metricsService.countName).To(BeIdenticalTo(""))
	})

	It("reports underlying errors to caller", func() {
		lct := &devicelastcontacttime.DeviceLastContactTime{LastContactTimestamp: 1337, DeviceId: 31337, OrgKey: "ORGKEY", DeregisterAfterTime: 30}
		marshalledLct, _ := proto.Marshal(lct)

		mockUnderlyingPublisher.putError = errors.New("SOME ERROR")
		metricsService.wg.Add(1)
		Expect(publisher.Publish(lct)).To(BeIdenticalTo(mockUnderlyingPublisher.putError))

		// We need to wait here to ensure the go routine that is counting errors gets a chance to run and count this one.
		metricsService.wg.Wait()

		Expect(mockUnderlyingPublisher.putCalled).To(BeComparableTo(true))
		Expect(mockUnderlyingPublisher.putCalledWithData).To(BeComparableTo(marshalledLct))
		Expect(mockUnderlyingPublisher.putCalledWithKey).To(BeComparableTo("31337"))
		Expect(metricsService.counter).To(BeIdenticalTo(int64(1)))
		Expect(metricsService.countName).To(BeIdenticalTo("kpl.hello-lct-stream.failed"))
	})

	AfterAll(func() {
		publisher.Stop()
	})
})

type logCall struct {
	msg    string
	fields []zapcore.Field
}
type mockLogger struct {
	calls map[string][]logCall
}

func newMockLogger() *mockLogger {
	return &mockLogger{calls: make(map[string][]logCall)}
}

// Debug implements logger.Logger.
func (m *mockLogger) Debug(msg string, fields ...zapcore.Field) {
	m.calls["Debug"] = append(m.calls["Debug"], logCall{msg: msg, fields: fields})
}

// Error implements logger.Logger.
func (m *mockLogger) Error(msg string, fields ...zapcore.Field) {
	m.calls["Error"] = append(m.calls["Error"], logCall{msg: msg, fields: fields})
}

// Info implements logger.Logger.
func (m *mockLogger) Info(string, ...zapcore.Field) {
	panic("unimplemented")
}

// Warn implements logger.Logger.
func (m *mockLogger) Warn(string, ...zapcore.Field) {
	panic("unimplemented")
}

// With implements logger.Logger.
func (m *mockLogger) With(...zapcore.Field) *zap.Logger {
	panic("unimplemented")
}

var _ = Describe("lct kinesis producer logger tests", func() {
	mockLogger := newMockLogger()
	testLogger := NewProducerLogger(mockLogger)
	It("logs info calls at debug to supress them", func() {
		testLogger.Info("info", producer.LogValue{Name: "foo", Value: "bar"})
		logCalls := mockLogger.calls["Debug"]
		Expect(len(logCalls)).To(BeEquivalentTo(1))
		logCall := logCalls[0]
		Expect(logCall.msg).To(BeComparableTo("info"))
		Expect(logCall.fields).To(BeComparableTo([]zapcore.Field{zap.Any("foo", "bar")}))
	})
	It("logs error calls at error", func() {
		testErr := errors.New("testError")
		testLogger.Error("error", testErr, producer.LogValue{Name: "foo", Value: "bar"})
		logCalls := mockLogger.calls["Error"]
		Expect(len(logCalls)).To(BeEquivalentTo(1))
		logCall := logCalls[0]
		Expect(len(logCall.fields)).To(BeEquivalentTo(2))
		Expect(logCall.msg).To(BeComparableTo("error"))
		Expect(logCall.fields[0]).To(BeComparableTo(zap.Any("foo", "bar")))
	})
})
