// Copyright (c) 2023 VMware Carbon Black, all rights reserved

// Package logger -- the logging wrapper for the hello service
package logger

import (
	"log"
	"sync"
	"time"

	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
	"hello-api/internal/config"
)

type (
	// Logger is an interface for a logger that mimics Zap level log functions
	// Used to pass both base logger levels + additional fields provided with `WithFields`
	Logger interface {
		Debug(string, ...zap.Field)
		Info(string, ...zap.Field)
		Warn(string, ...zap.Field)
		Error(string, ...zap.Field)

		With(...zap.Field) *zap.Logger
	}
)

var (
	//nolint:gochecknoglobals // needed for singleton
	baseLogger Logger
	logLevel   zap.AtomicLevel
	//nolint:gochecknoglobals // needed for singleton
	loggerOnce sync.Once
)

// BaseLogger returns a base logger for non-request related logging
func BaseLogger() Logger {
	loggerOnce.Do(func() {
		baseLogger, logLevel = newLoggerWithlevel("")
	})
	return baseLogger
}

func LogLevel() zap.AtomicLevel {
	return logLevel
}

func initLoggerConfig(cfg *config.LoggerConfig, out string) zap.Config {
	var lCfg zap.Config
	lCfg.DisableStacktrace = true
	if config.GetInstance().IsLocalEnv() {
		lCfg = zap.NewDevelopmentConfig()
	} else {
		lCfg = zap.NewProductionConfig()
	}
	if err := lCfg.Level.UnmarshalText([]byte(cfg.LogLevel)); err != nil {
		// must use built-in logger until base logger is initialized
		log.Println("unable to parse log level:", err)
		lCfg.Level = zap.NewAtomicLevelAt(zapcore.ErrorLevel)
	}
	if len(out) > 0 {
		lCfg.OutputPaths = []string{out}
	}
	lCfg.EncoderConfig.TimeKey = "timestamp"
	lCfg.EncoderConfig.EncodeTime = func(t time.Time, enc zapcore.PrimitiveArrayEncoder) {
		enc.AppendString(t.Format("2006-01-02 15:04:05.000000"))
	}
	return lCfg
}

func initLoggerOpts(cfg *config.LoggerConfig) []zap.Option {
	var opts []zap.Option
	if cfg.CommitSHA != "" {
		opts = append(opts, zap.Fields(zap.String("commit_sha", cfg.CommitSHA)))
	}
	return opts
}

// NewLogger returns a Logger
func NewLogger(out string) Logger {
	logger, _ := newLoggerWithlevel(out)
	return logger
}

func newLoggerWithlevel(out string) (Logger, zap.AtomicLevel) {
	cfg := config.GetLoggerConfigInstance()
	lCfg := initLoggerConfig(cfg, out)
	opts := initLoggerOpts(cfg)
	l, err := lCfg.Build(opts...)
	if err != nil {
		log.Fatalln("unable to build logger:", err)
	}

	return l, lCfg.Level
}
