// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package logger_test

import (
	"bytes"
	"net/url"
	"testing"
	"time"

	"hello-api/internal/config"
	"hello-api/internal/logger"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"go.uber.org/zap"
)

// MemorySink implements zap.Sink by writing all messages to a sink buffer.
type MemorySink struct {
	*bytes.Buffer
}

func (s *MemorySink) Close() error { return nil }
func (s *MemorySink) Sync() error  { return nil }

func TestRequestLogger(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Logger")
}

var cfg *config.Config

var _ = BeforeSuite(func() {
	cfg = config.GetInstance()
	cfg.EnvCategory = "local"
})

var _ = Describe("Request Logger", func() {
	var (
		requestLogger logger.Logger
		sink          *MemorySink
	)

	const (
		arg1    = "hello"
		arg2    = "goodbye"
		keyCS   = "commit_sha"
		valueCS = "abcdefghijklmnopqrstuvwxyz"
	)

	BeforeEach(func() {
		sink = &MemorySink{new(bytes.Buffer)}
		err := zap.RegisterSink("memory", func(*url.URL) (zap.Sink, error) {
			return sink, nil
		})
		if err != nil && err.Error() != "sink factory already registered for scheme \"memory\"" {
			Expect(err).To(BeNil())
		}
	})

	AfterEach(func() {
		cfg := config.GetLoggerConfigInstance()
		cfg.CommitSHA = ""
		cfg.LogLevel = "info"
	})

	Describe("Initialize logger", func() {
		Context("Creates a new request logger", func() {
			It("should create a valid logger with a commit SHA config and default level", func() {
				cfg := config.GetLoggerConfigInstance()
				cfg.CommitSHA = valueCS
				requestLogger = logger.NewLogger("memory://")

				requestLogger.Info("")

				Expect(sink.Buffer.String()).To(ContainSubstring(keyCS))
				Expect(sink.Buffer.String()).To(ContainSubstring(valueCS))
			})

			It("should create a valid logger with log level error as the default", func() {
				cfg := config.GetLoggerConfigInstance()
				cfg.LogLevel = "abcdefghijklmnop"
				requestLogger = logger.NewLogger("memory://")

				requestLogger.Debug("")
				requestLogger.Info("")
				requestLogger.Warn("")
				requestLogger.Error("")

				Expect(sink.Buffer.String()).ToNot(ContainSubstring("DEBUG"))
				Expect(sink.Buffer.String()).ToNot(ContainSubstring("INFO"))
				Expect(sink.Buffer.String()).ToNot(ContainSubstring("WARN"))
				Expect(sink.Buffer.String()).To(ContainSubstring("ERROR"))
			})
		})
	})

	Describe("Call logger out functions", func() {
		BeforeEach(func() {
			requestLogger = logger.NewLogger("memory://")
		})

		Context("All logger level interface functions", func() {
			It("should call Debug with non-empty message", func() {
				cfg := config.GetLoggerConfigInstance()
				cfg.LogLevel = "debug"
				requestLogger = logger.NewLogger("memory://")

				requestLogger.Debug("", zap.String(arg1, arg2))

				Expect(sink.Buffer.String()).To(ContainSubstring("DEBUG"))
				Expect(sink.Buffer.String()).To(ContainSubstring(`"` + arg1 + `": "` + arg2 + `"`))
			})

			It("should not call Debug because of default log level", func() {
				requestLogger.Debug("", zap.String(arg1, arg2))

				Expect(sink.Buffer.String()).ToNot(ContainSubstring(arg1))
				Expect(sink.Buffer.String()).ToNot(ContainSubstring(arg2))
			})

			It("should call Info with empty message", func() {
				requestLogger.Info("Message")

				Expect(sink.Buffer.String()).To(ContainSubstring("INFO"))
				Expect(sink.Buffer.String()).To(ContainSubstring("Message"))
			})

			It("should call Info non-empty message", func() {
				requestLogger.Info("", zap.String(arg1, arg2))

				Expect(sink.Buffer.String()).To(ContainSubstring("INFO"))
				Expect(sink.Buffer.String()).To(ContainSubstring(`"` + arg1 + `": "` + arg2 + `"`))
			})

			It("should call Warn", func() {
				requestLogger.Warn("", zap.String(arg1, arg2))

				Expect(sink.Buffer.String()).To(ContainSubstring("WARN"))
				Expect(sink.Buffer.String()).To(ContainSubstring(`"` + arg1 + `": "` + arg2 + `"`))
			})

			It("should call Error", func() {
				requestLogger.Error("", zap.String(arg1, arg2))

				Expect(sink.Buffer.String()).To(ContainSubstring("ERROR"))
				Expect(sink.Buffer.String()).To(ContainSubstring(`"` + arg1 + `": "` + arg2 + `"`))
			})
		})

		Context("Setting Fields", func() {
			It("should return an entry no fields", func() {
				requestLogger.Info("")

				// This date is magical: https://stackoverflow.com/questions/42217308/go-time-format-how-to-understand-meaning-of-2006-01-02-layout
				// Change at your own risk as the test will fail!
				expectedDate := time.Now().Format("2006-01-02")
				Expect(sink.Buffer.String()).To(ContainSubstring(expectedDate))
				Expect(sink.Buffer.String()).To(MatchRegexp("\\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1]) (2[0-3]|[01][0-9]):[0-5][0-9]:[0-5][0-9].\\d{6}"))
				Expect(sink.Buffer.String()).To(ContainSubstring("INFO"))
			})

			It("should return an entry with local fields", func() {
				requestLogger.With(zap.String(arg1, arg2)).Info("")

				Expect(sink.Buffer.String()).To(ContainSubstring(arg1 + `": "` + arg2))
			})

			It("should return an entry with local and base fields", func() {
				cfg := config.GetLoggerConfigInstance()
				cfg.CommitSHA = valueCS
				requestLogger = logger.NewLogger("memory://")
				requestLogger.With(zap.String(arg1, arg2)).Info("")

				Expect(sink.Buffer.String()).To(ContainSubstring(arg1 + `": "` + arg2))
				Expect(sink.Buffer.String()).To(ContainSubstring(keyCS + `": "` + valueCS))
			})

			It("should return an entry with local and base fields then just base fields", func() {
				cfg := config.GetLoggerConfigInstance()
				cfg.CommitSHA = valueCS
				requestLogger = logger.NewLogger("memory://")
				requestLogger.With(zap.String(arg1, arg2)).Info("")

				Expect(sink.Buffer.String()).To(ContainSubstring(arg1 + `": "` + arg2))
				Expect(sink.Buffer.String()).To(ContainSubstring(keyCS + `": "` + valueCS))

				sink.Reset()
				requestLogger.Info("")

				Expect(sink.Buffer.String()).ToNot(ContainSubstring(arg1 + `": "` + arg2))
				Expect(sink.Buffer.String()).To(ContainSubstring(keyCS + `": "` + valueCS))
			})

			It("should return an entry with local and base fields, and base fields are overwritten by local", func() {
				cfg := config.GetLoggerConfigInstance()
				cfg.CommitSHA = valueCS
				requestLogger = logger.NewLogger("memory://")
				requestLogger.With(zap.String(arg1, arg2)).Info("")

				Expect(sink.Buffer.String()).To(ContainSubstring(arg1 + `": "` + arg2))
				Expect(sink.Buffer.String()).To(ContainSubstring(keyCS + `": "` + valueCS))
			})

			It("should set a field on the logger", func() {
				requestLogger.With(zap.String(arg1, arg2)).Info("")

				Expect(sink.Buffer.String()).To(ContainSubstring(arg1 + `": "` + arg2))
			})
		})
	})
})
