// Copyright (c) 2023 VMware Carbon Black, all rights reserved

// Package logger -- the logging wrapper for the hello service
package logger

import (
	"errors"
	"sync"

	"hello-api/internal/config"
	"hello-api/internal/logger/s3logger"

	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
)

type (
	StoppableLogger interface {
		Logger
		Stop() error
	}
)

var (
	//nolint:gochecknoglobals // needed for singleton
	s3Logger FilteredLogger
	//nolint:gochecknoglobals // needed for singleton
	s3LoggerOnce sync.Once
)

func S3Logger() FilteredLogger {
	s3LoggerOnce.Do(func() {
		s3Logger = news3Logger()
	})
	return s3Logger
}

func news3Logger() FilteredLogger {
	cfg := config.GetLoggerConfigInstance()
	if cfg.UseS3 {
		wholeConfig := config.GetInstance()
		s3Writer, err := s3logger.NewS3WriteSyncer(cfg.S3Bucket, cfg.S3ChunkSize, wholeConfig)
		if err != nil {
			BaseLogger().Warn("Failed to enable s3 logger", zap.Error(err))
			return &nopBloomedLogger{}
		}
		s3Core := zapcore.NewCore(zapcore.NewJSONEncoder(zap.NewProductionEncoderConfig()), zapcore.Lock(s3Writer), zap.LevelEnablerFunc(func(lvl zapcore.Level) bool {
			return true
		}))
		BaseLogger().Warn("Enabling s3 logger", zap.String("bucket", cfg.S3Bucket))
		logger, err := NewBloomedLogger(stopLoggerFunc(zap.New(s3Core), s3Writer.Stop))
		if err != nil {
			BaseLogger().Warn("Failed to enable s3 logger", zap.Error(err))
			return newNopBloomLogger()
		}
		return logger
	}
	return newNopBloomLogger()
}

func nopStopLogger(l *zap.Logger) StoppableLogger {
	return &stopableLogger{Logger: l, onStop: func() error { return nil }}
}

func stopLoggerFunc(l *zap.Logger, onStop func() error) StoppableLogger {
	return &stopableLogger{Logger: l, onStop: onStop}
}

type stopableLogger struct {
	*zap.Logger
	onStop func() error
}

func (stopLogger *stopableLogger) Stop() error {
	syncErr := stopLogger.Sync()
	stopErr := stopLogger.onStop()
	return errors.Join(syncErr, stopErr)
}
