package s3logger

import (
	"fmt"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

type mockWriteCloser struct {
	onWrite  func([]byte) (int, error)
	onClose  func() error
	isClosed bool
}

func (m *mockWriteCloser) Write(data []byte) (int, error) {
	return m.onWrite(data)
}

func (m *mockWriteCloser) Close() error {
	m.isClosed = true
	return m.onClose()
}

func mockWriteCloserFunc(closeError, writeError error) *mockWriteCloser {
	return &mockWriteCloser{
		onWrite: func(data []byte) (int, error) {
			return len(data), writeError
		},
		onClose: func() error {
			return closeError
		},
	}
}

func mockWriteCloserGood() *mockWriteCloser {
	return mockWriteCloserFunc(nil, nil)
}

var _ = Describe("s3 limited writer", func() {
	Context("limited writer ", func() {
		It("writes correctly", func() {
			mockWriter := mockWriteCloserGood()
			testChunk := limitedWriteCloser{
				limit:  10,
				Writer: mockWriter,
			}
			testData := []byte{1, 2, 3, 4}
			written, err := testChunk.Write(testData)
			Expect(err).To(BeNil())
			Expect(written).To(BeEquivalentTo(4))
			Expect(mockWriter.isClosed).To(BeFalse())
		})
		It("handles underlying writer errors correctly on write", func() {
			mockWriter := mockWriteCloserFunc(nil, fmt.Errorf("Some error"))
			testChunk := limitedWriteCloser{
				limit:  10,
				Writer: mockWriter,
			}
			testData := []byte{1, 2, 3, 4}
			_, err := testChunk.Write(testData)
			Expect(err).ToNot(BeNil())
		})
		It("writes correctly up to and past the limit", func() {
			mockWriter := mockWriteCloserGood()

			testChunk := limitedWriteCloser{
				limit:  10,
				Writer: mockWriter,
			}
			testData := [11]byte{}
			written, err := testChunk.Write(testData[:])
			Expect(err).To(MatchError(ErrLimitHit))
			Expect(written).To(BeEquivalentTo(written))
		})
		It("closes", func() {
			mockWriter := mockWriteCloserGood()
			testChunk := limitedWriteCloser{
				limit:  10,
				Writer: mockWriter,
			}
			err := testChunk.Close()
			Expect(err).To(BeNil())
			Expect(mockWriter.isClosed).To(BeTrue())
		})
	})
})
