package s3logger

import (
	"context"
	"errors"
	"fmt"
	"io"
	"time"

	"hello-api/internal/config"

	"github.com/aws/aws-sdk-go-v2/aws"
	awsconfig "github.com/aws/aws-sdk-go-v2/config"
	"github.com/aws/aws-sdk-go-v2/feature/s3/manager"
	"github.com/aws/aws-sdk-go-v2/service/s3"
	"go.uber.org/zap/zapcore"
)

type (
	StopableWriteSyncer interface {
		zapcore.WriteSyncer // interface for zapcore Writers that can Sync
		Stop() error        // stop to cleanup resources
	}

	// Note: not go-routine safe, use zapcore.Lock() to add mutex
	s3UploadWriter struct {
		uploadSubmitter s3UploadSubmitter
		currentUpload   io.WriteCloser
	}
)

func NewS3WriteSyncer(bucketName string, chunkSize int, conf *config.Config) (StopableWriteSyncer, error) {
	awsConfig, err := awsconfig.LoadDefaultConfig(context.Background(),
		awsconfig.WithRegion(conf.Region),
	)
	if err != nil {
		return nil, err
	}
	client := s3.NewFromConfig(awsConfig, func(o *s3.Options) {
		if conf.IsLocalEnv() {
			o.UsePathStyle = true
			o.BaseEndpoint = &conf.AwsEndpoint
			o.EndpointOptions.DisableHTTPS = true
			o.EndpointOptions.UseDualStackEndpoint = aws.DualStackEndpointStateDisabled
		}
	})

	timedCtx, cancelFunc := context.WithTimeout(context.Background(), 2*time.Second)
	defer cancelFunc()
	_, err = client.HeadBucket(timedCtx, &s3.HeadBucketInput{Bucket: aws.String(bucketName)})
	if err != nil {
		return nil, fmt.Errorf("bucket %s does not exist or is not accesible: %w", bucketName, err)
	}

	uploader := manager.NewUploader(client)
	pooledUploader := NewPooledS3UploadSubmitter(bucketName, chunkSize, uploader)
	return newS3Writer(pooledUploader)
}

func newS3Writer(s3Uploader s3UploadSubmitter) (*s3UploadWriter, error) {
	uploadWriter := &s3UploadWriter{uploadSubmitter: s3Uploader}
	return uploadWriter, uploadWriter.newUpload()
}

// Note: not go-routine safe, use zapcore.Lock() to add mutex
func (w *s3UploadWriter) Write(data []byte) (n int, err error) {
	n, err = w.currentUpload.Write(data)
	if errors.Is(err, ErrLimitHit) {
		err = w.newUpload()
	}
	return n, err
}

func (w *s3UploadWriter) Sync() error {
	return w.newUpload()
}

func (w *s3UploadWriter) Stop() (err error) {
	if w.currentUpload != nil {
		err = w.currentUpload.Close()
	}
	w.uploadSubmitter.Stop()
	return err
}

func (w *s3UploadWriter) newUpload() (err error) {
	if w.currentUpload != nil {
		err = w.currentUpload.Close()
	}

	w.currentUpload = w.uploadSubmitter.SubmitUpload()
	return err
}
