package s3logger

import (
	"context"
	"io"
	"testing"

	"github.com/aws/aws-sdk-go-v2/feature/s3/manager"
	"github.com/aws/aws-sdk-go-v2/service/s3"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

func TestS3Writer(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "s3 writing utils")
}

type mockSyncS3UploadSubmitter struct {
	uploader      s3Uploader
	uploads       []*s3.PutObjectInput
	currentWriter io.WriteCloser
}

func (mock *mockSyncS3UploadSubmitter) SubmitUpload() io.WriteCloser {
	//nolint:errcheck
	mock.uploader.Upload(context.TODO(), &s3.PutObjectInput{})
	mock.uploads = append(mock.uploads, &s3.PutObjectInput{})
	return mock.currentWriter
}

func (mock *mockSyncS3UploadSubmitter) Stop() {
}

type mockS3Uploader struct {
	uploadError error
	input       *s3.PutObjectInput
}

func (ms3u *mockS3Uploader) Upload(ctx context.Context, input *s3.PutObjectInput, opts ...func(*manager.Uploader)) (
	*manager.UploadOutput, error,
) {
	ms3u.input = input
	return nil, ms3u.uploadError
}

var _ = Describe("s3 write syncer", func() {
	var testWriter *s3UploadWriter
	var err error
	var mockUploadSubmitter *mockSyncS3UploadSubmitter

	BeforeEach(func() {
		goodMockWriter := mockWriteCloserGood()
		mockUploader := &mockS3Uploader{}
		mockUploadSubmitter = &mockSyncS3UploadSubmitter{uploader: mockUploader, currentWriter: goodMockWriter}
		testWriter, err = newS3Writer(mockUploadSubmitter)
		Expect(err).To(BeNil())
		Expect(testWriter.currentUpload).ToNot(BeNil())
		Expect(len(mockUploadSubmitter.uploads)).To(BeComparableTo(1))
	})

	Context("s3 writer", func() {
		It("writes correctly", func() {
			testData := [5]byte{}
			n, err := testWriter.Write(testData[:])
			Expect(err).To(BeNil())
			Expect(n).To(BeEquivalentTo(5))
			Expect(testWriter.currentUpload).ToNot(BeNil())
			Expect(len(mockUploadSubmitter.uploads)).To(BeComparableTo(1))
		})
		It("writes correctly past limit", func() {
			goodMockWriter := mockWriteCloserFunc(nil, ErrLimitHit)
			mockUploader := &mockS3Uploader{}
			mockUploadSubmitter = &mockSyncS3UploadSubmitter{uploader: mockUploader, currentWriter: goodMockWriter}
			testWriter, err = newS3Writer(mockUploadSubmitter)
			testData := [20]byte{}
			n, err := testWriter.Write(testData[:])
			Expect(err).To(BeNil())
			Expect(n).To(BeEquivalentTo(20))
			Expect(testWriter.currentUpload).ToNot(BeNil())
			Expect(len(mockUploadSubmitter.uploads)).To(BeComparableTo(2))
		})
		It("syncs correctly", func() {
			err := testWriter.Sync()
			Expect(err).To(BeNil())
			Expect(testWriter.currentUpload).ToNot(BeNil())
			Expect(len(mockUploadSubmitter.uploads)).To(BeComparableTo(2))
		})
		It("Closes correctly", func() {
			err := testWriter.Stop()
			Expect(err).To(BeNil())
			Expect(testWriter.currentUpload).ToNot(BeNil())
		})
		It("generates new uploads correctly", func() {
			err := testWriter.newUpload()
			Expect(err).To(BeNil())
			Expect(len(mockUploadSubmitter.uploads)).To(BeComparableTo(2))
		})
	})
})
