package s3logger

import (
	"bufio"
	"context"
	"fmt"
	"io"
	"log"
	"time"

	"github.com/alitto/pond"
	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/feature/s3/manager"
	"github.com/aws/aws-sdk-go-v2/service/s3"
)

type (
	s3Uploader interface {
		Upload(ctx context.Context, input *s3.PutObjectInput, opts ...func(*manager.Uploader)) (
			*manager.UploadOutput, error,
		)
	}

	taskSubmitter interface {
		Submit(task func())
		StopAndWaitFor(time.Duration)
	}

	s3UploadSubmitter interface {
		SubmitUpload() io.WriteCloser
		Stop()
	}

	uploadTaskSubmitter struct {
		taskSubmitter taskSubmitter
		cancelFunc    context.CancelFunc
		taskContext   context.Context
		uploader      s3Uploader
		limit         int
		bucketName    string
	}
)

func NewPooledS3UploadSubmitter(bucketName string, limit int, uploader s3Uploader) *uploadTaskSubmitter {
	pool := pond.New(5, 2)
	return newS3UploadSubmitter(bucketName, limit, pool, uploader)
}

func newS3UploadSubmitter(bucketName string, limit int, taskSubmitter taskSubmitter, uploader s3Uploader) *uploadTaskSubmitter {
	taskContext, taskCancelFunc := context.WithCancel(context.Background())
	return &uploadTaskSubmitter{bucketName: bucketName, limit: limit, taskSubmitter: taskSubmitter, taskContext: taskContext, uploader: uploader, cancelFunc: taskCancelFunc}
}

func (u *uploadTaskSubmitter) SubmitUpload() io.WriteCloser {
	// the reader from io.Pipe is explicitly go-routine safe
	pr, pw := io.Pipe()

	upParams := &s3.PutObjectInput{
		Bucket: aws.String(u.bucketName),
		Key:    aws.String(fmt.Sprintf("hello-logs-%d.json", time.Now().UnixMilli())),
		Body:   pr,
	}

	u.taskSubmitter.Submit(func() {
		_, err := u.uploader.Upload(u.taskContext, upParams)
		if err != nil {
			log.Println(fmt.Errorf("Error uploading to %s: %w", u.bucketName, err))
		}
	})

	upload := limitedWriteCloser{
		Writer: newBufferedWriteCloser(pw),
		limit:  u.limit,
	}

	return &upload
}

func (u *uploadTaskSubmitter) Stop() {
	u.taskSubmitter.StopAndWaitFor(5 * time.Second)
	u.cancelFunc()
}

func newBufferedWriteCloser(underylingWriter io.WriteCloser) io.WriteCloser {
	return &writeCloser{underlyingWriter: underylingWriter, Writer: bufio.NewWriter(underylingWriter)}
}

type writeCloser struct {
	underlyingWriter io.WriteCloser
	*bufio.Writer
}

func (mwc *writeCloser) Close() error {
	if err := mwc.Flush(); err != nil {
		return err
	}
	return mwc.underlyingWriter.Close()
}
