// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package metrics

import (
	"hello-api/internal/metrics/prometheus"
	"time"

	"hello-api/internal/config"
	"hello-api/internal/metrics/wavefront"

	"github.com/opentracing/opentracing-go"
)

// Service includes metric gathering functions to be called regardless of client impl
type Service interface {
	TrackMethodTime(name string, start time.Time, dims map[string]string)

	Count(name string, dimensions map[string]string, value int64)
	Gauge(name string, dimensions map[string]string, value int64)
	Histogram(name string, dimensions map[string]string, value int64)
	RegisterMetric(name string, metric interface{}, tags map[string]string) error

	Tracer() opentracing.Tracer
	CloseConnection()

	Name() string
}

type Clients struct {
	reporters []Service
}

// NewClient returns a metrics client if configured to use one
func NewClient(cfg *config.Config) (Service, error) {
	return NewClients(cfg)
}

func NewClients(cfg *config.Config) (Service, error) {
	var reporters []Service

	if cfg.WavefrontEnabled {
		if wfReporter, wfErr := wavefront.NewReporter(cfg); wfErr != nil {
			return &Clients{
				reporters: reporters,
			}, wfErr
		} else {
			reporters = append(reporters, wfReporter)
		}
	}

	if cfg.PrometheusEnabled {
		reporters = append(reporters, prometheus.NewReporter())
	}

	return &Clients{
		reporters: reporters,
	}, nil
}

func (c Clients) TrackMethodTime(name string, start time.Time, dims map[string]string) {
	for _, r := range c.reporters {
		r.TrackMethodTime(name, start, dims)
	}
}

func (c Clients) Count(name string, dimensions map[string]string, value int64) {
	for _, r := range c.reporters {
		r.Count(name, dimensions, value)
	}
}

func (c Clients) Gauge(name string, dimensions map[string]string, value int64) {
	for _, r := range c.reporters {
		r.Gauge(name, dimensions, value)
	}
}

func (c Clients) Histogram(name string, dimensions map[string]string, value int64) {
	for _, r := range c.reporters {
		r.Histogram(name, dimensions, value)
	}
}

func (c Clients) RegisterMetric(name string, metric interface{}, tags map[string]string) error {
	var e error
	for _, r := range c.reporters {
		e = r.RegisterMetric(name, metric, tags)
	}

	return e
}

// Tracer isn't supported by Prometheus, so we only care about returning Wavefront in this case
func (c Clients) Tracer() opentracing.Tracer {
	for _, r := range c.reporters {
		if r.Name() == wavefront.Name {
			return r.Tracer()
		}
	}

	return nil
}

func (c Clients) CloseConnection() {
	for _, r := range c.reporters {
		r.CloseConnection()
	}
}

func (c Clients) Name() string {
	var name string
	for _, r := range c.reporters {
		name += r.Name()
	}

	return name
}
