// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package metrics

import (
	"fmt"
	"runtime"

	"hello-api/internal/logger"

	gometrics "github.com/rcrowley/go-metrics"
)

// goStat is an enum for the type of Redis stat
type goStat string

const (
	alloc        goStat = "alloc"
	totalAlloc   goStat = "totalAlloc"
	sys          goStat = "sys"
	lookups      goStat = "lookups"
	mallocs      goStat = "mallocs"
	frees        goStat = "frees"
	heapAlloc    goStat = "heapAlloc"
	heapSys      goStat = "heapSys"
	heapIdle     goStat = "heapIdle"
	heapInUse    goStat = "heapInUse"
	heapReleased goStat = "heapReleased"
	heapObjects  goStat = "heapObjects"
	stackInUse   goStat = "stackInUse"
	stackSys     goStat = "stackSys"
	mSpanInUse   goStat = "mSpanInUse"
	mSpanSys     goStat = "mSpanSys"
	mCacheInUse  goStat = "mCacheInUse"
	mCacheSys    goStat = "mCacheSys"
	buckHashSys  goStat = "buckHashSys"
	gcSys        goStat = "gcSys"
	otherSys     goStat = "otherSys"
	pauseTotalNS goStat = "pauseTotalNS"
	nextGC       goStat = "nextGC"
	lastGC       goStat = "lastGC"
	numGC        goStat = "numGC"
	numCGOCall   goStat = "numCGOCall"
	numGoRoutine goStat = "numGoRoutine"
)

// goStatGaugeMetric
type goStatGaugeMetric struct {
	stat goStat
}

//nolint:gocyclo,funlen // despite high cyclo complexity and length, this function is very straightforward and would not be served well to be broken down
func (m *goStatGaugeMetric) Value() int64 {
	stats := runtime.MemStats{}
	runtime.ReadMemStats(&stats)
	switch m.stat {
	case alloc:
		return int64(stats.Alloc)
	case totalAlloc:
		return int64(stats.TotalAlloc)
	case sys:
		return int64(stats.Sys)
	case lookups:
		return int64(stats.Lookups)
	case mallocs:
		return int64(stats.Mallocs)
	case frees:
		return int64(stats.Frees)
	case heapAlloc:
		return int64(stats.HeapAlloc)
	case heapSys:
		return int64(stats.HeapSys)
	case heapIdle:
		return int64(stats.HeapIdle)
	case heapInUse:
		return int64(stats.HeapInuse)
	case heapReleased:
		return int64(stats.HeapReleased)
	case heapObjects:
		return int64(stats.HeapObjects)
	case stackInUse:
		return int64(stats.StackInuse)
	case stackSys:
		return int64(stats.StackSys)
	case mSpanInUse:
		return int64(stats.MSpanInuse)
	case mSpanSys:
		return int64(stats.MSpanSys)
	case mCacheInUse:
		return int64(stats.MCacheInuse)
	case mCacheSys:
		return int64(stats.MCacheSys)
	case buckHashSys:
		return int64(stats.BuckHashSys)
	case gcSys:
		return int64(stats.GCSys)
	case otherSys:
		return int64(stats.OtherSys)
	case pauseTotalNS:
		return int64(stats.PauseTotalNs)
	case nextGC:
		return int64(stats.NextGC)
	case lastGC:
		return int64(stats.LastGC)
	case numGC:
		return int64(stats.NumGC)
	case numCGOCall:
		return runtime.NumCgoCall()
	case numGoRoutine:
		return int64(runtime.NumGoroutine())
	default:
		return 0
	}
}

func (m *goStatGaugeMetric) Update(_ int64) {
	panic("update not supported")
}

func (m *goStatGaugeMetric) Snapshot() gometrics.Gauge {
	return gometrics.GaugeSnapshot(m.Value())
}

// RegisterGoStatMetrics registers Wavefront metrics for each Redis stat
//   - NOTE: if we have metrics errors, flag it with an error in the log but continue onward
func RegisterGoStatMetrics(service Service) {
	gaugeStats := []goStat{
		alloc,
		totalAlloc,
		sys,
		lookups,
		mallocs,
		frees,
		heapAlloc,
		heapSys,
		heapIdle,
		heapInUse,
		heapReleased,
		heapObjects,
		stackInUse,
		stackSys,
		mSpanInUse,
		mSpanSys,
		mCacheInUse,
		mCacheSys,
		buckHashSys,
		gcSys,
		otherSys,
		pauseTotalNS,
		nextGC,
		lastGC,
		numGC,
		numCGOCall,
		numGoRoutine,
	}

	tags := map[string]string{}
	for _, stat := range gaugeStats {
		if err := service.RegisterMetric("go."+string(stat), &goStatGaugeMetric{stat: stat}, tags); err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("error registering metric: %v", stat))
		}
	}
}
