package prometheus

import (
	"fmt"
	"io"
	"net/http"
	"os"
	"strings"
	"time"

	"hello-api/internal/config"
	"hello-api/internal/logger"

	"github.com/opentracing/opentracing-go"
	prometheusClient "github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	jconfig "github.com/uber/jaeger-client-go/config"
	"golang.org/x/exp/maps"
)

const Name = "PROMETHEUS"
const prefix = "ems_device_hello_"

type metricType int

const (
	// Counter metric from go-metrics
	Counter metricType = iota
	// Gauge metric
	Gauge
	// Histogram metric
	Histogram
)

type Reporter struct {
	reporter                prometheusClient.Registerer
	tracerClose             io.Closer
	tracer                  opentracing.Tracer
	metrics                 map[string]interface{}
	metricsToUpdateOnScrape map[string]interface{}
	name                    string
}

func NewReporter() *Reporter {
	conf := jconfig.Configuration{
		ServiceName: "ems-device-hello",
		Sampler: &jconfig.SamplerConfig{
			Type:  "const",
			Param: 1,
		},
		Reporter: &jconfig.ReporterConfig{
			LogSpans: true,
		},
	}

	tracer, closer, err := conf.NewTracer()
	if err != nil {
		logger.BaseLogger().Warn("Couldn't open tracer for Prometheus, but doesn't matter.")
	}
	opentracing.SetGlobalTracer(tracer)

	cfg := config.GetInstance()

	registry := prometheusClient.WrapRegistererWithPrefix(prefix, prometheusClient.DefaultRegisterer)
	registry = prometheusClient.WrapRegistererWith(map[string]string{
		"env": cfg.CbdEnvName,
		"source": func() string {
			hostname, err := os.Hostname()
			if err != nil {
				hostname = "unknown"
			}

			return hostname
		}(),
	}, registry)

	prometheusClient.DefaultRegisterer = registry

	reporter := &Reporter{
		reporter:                registry,
		name:                    Name,
		tracer:                  tracer,
		tracerClose:             closer,
		metrics:                 make(map[string]interface{}),
		metricsToUpdateOnScrape: make(map[string]interface{}),
	}

	reporter.newPrometheusServer()

	return reporter
}

type StatMetric interface {
	Value() int64
}

func (r *Reporter) newPrometheusServer() {
	cfg := config.GetInstance()

	logger.BaseLogger().Info(fmt.Sprintf("Prometheus scrape endpoint on port %d", cfg.PrometheusPort))

	server := &http.Server{
		Addr:              fmt.Sprintf(":%d", cfg.PrometheusPort),
		ReadHeaderTimeout: 3 * time.Second,
	}

	go func() {
		http.Handle("/"+cfg.ServicePrefix+"/prometheus", r.statMetricUpdateMiddleware(promhttp.Handler()))
		err := server.ListenAndServe()
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Unable to start prometheus scrape endpoint %e", err))
		}
	}()
}

func (r *Reporter) statMetricUpdateMiddleware(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, request *http.Request) {
		for name, m := range r.metricsToUpdateOnScrape {
			r.Gauge(name, nil, m.(StatMetric).Value())
		}

		next.ServeHTTP(w, request)
	})
}

func (r *Reporter) RegisterMetric(name string, metric interface{}, tags map[string]string) error {
	prometheusCompliantName := strings.ReplaceAll(name, ".", "_")

	if _, ok := r.metricsToUpdateOnScrape[prometheusCompliantName]; ok {
		logger.BaseLogger().Warn(fmt.Sprintf("Dropping registration of metric %s as it has already been registered.", prometheusCompliantName))
		return fmt.Errorf("metric already registered")
	}

	// This is currently used by db_stat_metric, redis_stat_metric, and go_stat_metric which all use Gauges
	// We'll have to account for other metricTypes if we introduce a stat metric that uses another type
	m := prometheusClient.NewGaugeVec(prometheusClient.GaugeOpts{
		Name: prometheusCompliantName,
	}, nil)
	r.metricsToUpdateOnScrape[prometheusCompliantName] = metric
	r.metrics[prometheusCompliantName] = m
	r.reporter.MustRegister(m)

	return nil
}

func (r *Reporter) TrackMethodTime(name string, start time.Time, dims map[string]string) {
	r.Histogram(
		name+"_duration_micros",
		dims,
		time.Since(start).Nanoseconds()/1e3,
	)
}

func (r *Reporter) Count(name string, dimensions map[string]string, value int64) {
	metric := r.getNewOrExistingMetric(name, Counter, dimensions)
	if metric == nil {
		return
	}

	metric.(*prometheusClient.CounterVec).With(dimensions).Add(float64(value))
}

func (r *Reporter) Gauge(name string, dimensions map[string]string, value int64) {
	metric := r.getNewOrExistingMetric(name, Gauge, dimensions)
	if metric == nil {
		return
	}

	metric.(*prometheusClient.GaugeVec).With(dimensions).Set(float64(value))
}

func (r *Reporter) Histogram(name string, dimensions map[string]string, value int64) {
	metric := r.getNewOrExistingMetric(name, Histogram, dimensions)
	if metric == nil {
		return
	}

	metric.(*prometheusClient.HistogramVec).With(dimensions).Observe(float64(value))
}

func (r *Reporter) Tracer() opentracing.Tracer {
	return r.tracer
}

func (r *Reporter) CloseConnection() {
	if err := r.tracerClose.Close(); err != nil {
		logger.BaseLogger().Warn(fmt.Sprintf("Could not close tracer %v", err))
	}
}

func (r *Reporter) Name() string {
	return r.name
}

func (r *Reporter) getNewOrExistingMetric(name string, mt metricType, dimensions map[string]string) interface{} {
	prometheusCompliantName := strings.ReplaceAll(name, ".", "_")
	if metric, ok := r.metrics[prometheusCompliantName]; ok {
		return metric
	}

	switch mt {
	case Counter:
		metric := prometheusClient.NewCounterVec(prometheusClient.CounterOpts{
			Name: prometheusCompliantName,
		},
			maps.Keys(dimensions))
		r.metrics[prometheusCompliantName] = metric
		r.reporter.MustRegister(metric)
		return metric
	case Gauge:
		metric := prometheusClient.NewGaugeVec(prometheusClient.GaugeOpts{
			Name: prometheusCompliantName,
		},
			maps.Keys(dimensions))
		r.metrics[prometheusCompliantName] = metric
		r.reporter.MustRegister(metric)
		return metric
	case Histogram:
		metric := prometheusClient.NewHistogramVec(prometheusClient.HistogramOpts{},
			maps.Keys(dimensions))
		r.metrics[prometheusCompliantName] = metric
		r.reporter.MustRegister(metric)
		return metric
	default:
		logger.BaseLogger().Error("bad metric type")
		return nil
	}
}
