// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package wavefront

import (
	"fmt"

	"hello-api/internal/config"
	"hello-api/internal/logger"

	"github.com/opentracing/opentracing-go"
	gometrics "github.com/rcrowley/go-metrics"
	"github.com/wavefronthq/go-metrics-wavefront/reporting"
	"github.com/wavefronthq/wavefront-opentracing-sdk-go/reporter"
	"github.com/wavefronthq/wavefront-opentracing-sdk-go/tracer"
	"github.com/wavefronthq/wavefront-sdk-go/application"
	"github.com/wavefronthq/wavefront-sdk-go/senders"
)

// Reporter is the Wavefront facing struct that handles the metrics corresponding to the client
type Reporter struct {
	reporter reporting.WavefrontMetricsReporter
	tracer   opentracing.Tracer
	metrics  map[string]interface{}
	cfg      *config.Config
	name     string
}

type metricType int

const Name = "WAVEFRONT"

const (
	// Counter metric from go-metrics
	Counter metricType = iota
	// Gauge metric
	Gauge
	// Histogram metric
	Histogram
)

// NewReporter returns a wavefront reporter with received configuration and a metric interval
// Similar to the SignalFx Scheduler
func NewReporter(cfg *config.Config) (*Reporter, error) {
	logger.BaseLogger().Info(fmt.Sprintf("Wavefront reporter starting with scheduled interval time: %v", cfg.MetricInterval))
	sender, err := senders.NewSender(cfg.WavefrontEndpoint, senders.MetricsPort(cfg.WavefrontPort), senders.TracesPort(cfg.WavefrontTracePort))
	if err != nil {
		return nil, err
	}
	appTags := &application.Tags{
		Application: cfg.BaseDimensions["product"],
		Service:     cfg.BaseDimensions["service"],
		Cluster:     cfg.BaseDimensions["env"],
		Shard:       "none",
	}

	metricReporter := reporting.NewMetricsReporter(
		sender,
		reporting.ApplicationTag(*appTags),
		reporting.CustomRegistry(gometrics.NewRegistry()),
		reporting.Prefix(cfg.MetricPrefix),
		reporting.Interval(cfg.MetricInterval),
		reporting.LogErrors(true),
	)

	spanReporter := reporter.New(
		sender,
		*appTags,
		reporter.RedMetricsCustomTagKeys([]string{"http.status_code", "http.user_agent", "http.accept", "http.cache_hit"}),
		reporter.BufferSize(20000),
	)
	rateSampler := tracer.RateSampler{Rate: cfg.MetricSamplerRate}
	globalTracer := tracer.New(spanReporter, tracer.WithSampler(rateSampler))
	opentracing.InitGlobalTracer(globalTracer)
	return &Reporter{
		reporter: metricReporter,
		tracer:   globalTracer,
		metrics:  make(map[string]interface{}),
		cfg:      cfg,
		name:     Name,
	}, nil
}

// Tracer gets the reporter's tracer
func (c *Reporter) Tracer() opentracing.Tracer {
	return c.tracer
}

func (c *Reporter) GetConfig() *config.Config {
	return c.cfg
}

// CloseConnection will tell the reporter to gracefully close which in term will close the sender attached
func (c *Reporter) CloseConnection() {
	c.reporter.Close()
}

// Count finds the counter metric matching the provided name  in the reporter and increments
// If no Counter metric exists, will register a new counter
func (c *Reporter) Count(name string, dimensions map[string]string, value int64) {
	metric := c.getNewOrExistingMetric(name, Counter)
	if metric == nil {
		return
	}

	counter := c.reporter.GetOrRegisterMetric(name, metric, dimensions)
	counter.(gometrics.Counter).Inc(value)
}

// Gauge finds the gauge metric matching the provided name in the reporter and updates the value
// If no Gauge metric exists, will register a new gauge
func (c *Reporter) Gauge(name string, dimensions map[string]string, value int64) {
	metric := c.getNewOrExistingMetric(name, Gauge)
	if metric == nil {
		return
	}

	gauge := c.reporter.GetOrRegisterMetric(name, metric, dimensions)
	gauge.(gometrics.Gauge).Update(value)
}

// Histogram finds the histogram metric matching the provided name in the reporter and updates the distribution
// If no Histogram metric exists, will register a new histogram
func (c *Reporter) Histogram(name string, dimensions map[string]string, value int64) {
	metric := c.getNewOrExistingMetric(name, Histogram)
	if metric == nil {
		return
	}

	histogram := c.reporter.GetOrRegisterMetric(name, metric, dimensions)
	histogram.(gometrics.Histogram).Update(value)
}

// RegisterMetric registers a metric with the Wavefront reporter
func (c *Reporter) RegisterMetric(name string, metric interface{}, tags map[string]string) error {
	return c.reporter.RegisterMetric(name, metric, tags)
}

func (c *Reporter) Name() string {
	return c.name
}

func (c *Reporter) getNewOrExistingMetric(name string, mt metricType) interface{} {
	if metric, ok := c.metrics[name]; ok {
		return metric
	}

	switch mt {
	case Counter:
		metric := gometrics.NewCounter()
		c.metrics[name] = metric
		return metric
	case Gauge:
		metric := gometrics.NewGauge()
		c.metrics[name] = metric
		return metric
	case Histogram:
		metric := reporting.NewHistogram()
		c.metrics[name] = metric
		return metric
	default:
		logger.BaseLogger().Error("bad metric type")
		return nil
	}
}
