// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package streamchecker

import (
	"context"
	"fmt"
	"time"

	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/service/kinesis"
)

type (
	StreamChecker interface {
		DoesStreamExist(ctx context.Context) error
	}

	streamDescriber interface {
		DescribeStream(ctx context.Context, params *kinesis.DescribeStreamInput, optFns ...func(*kinesis.Options)) (*kinesis.DescribeStreamOutput, error)
	}

	streamChecker struct {
		streamName string
		client     streamDescriber
	}
)

func (checker *streamChecker) DoesStreamExist(ctx context.Context) error {
	describeStreamReq := &kinesis.DescribeStreamInput{
		ExclusiveStartShardId: nil,
		Limit:                 nil,
		StreamName:            aws.String(checker.streamName),
	}
	_, err := checker.client.DescribeStream(ctx, describeStreamReq)
	if err != nil {
		return fmt.Errorf("output lct stream '%s' cannot be connected:\n  %w", checker.streamName, err)
	}
	return nil
}

func NewStreamChecker(streamName string, kinesisClient streamDescriber) StreamChecker {
	return &streamChecker{streamName: streamName, client: kinesisClient}
}

func NewStreamCheckerWithTimeout(streamName string, kinesisClient streamDescriber, timeout time.Duration) StreamChecker {
	checker := NewStreamChecker(streamName, kinesisClient)
	return &timedStreamChecker{StreamChecker: checker, timeout: timeout}
}

type timedStreamChecker struct {
	timeout time.Duration
	StreamChecker
}

func (timedChecker *timedStreamChecker) DoesStreamExist(ctx context.Context) error {
	timedCtx, cancelFunc := context.WithTimeout(ctx, timedChecker.timeout)
	defer cancelFunc()
	return timedChecker.StreamChecker.DoesStreamExist(timedCtx)
}
