// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package utils

import (
	"errors"
	"fmt"
	"os"
	"strings"
	"time"

	"hello-api/internal/logger"

	"hello-api/internal/app/parsers/devicecontext"

	"github.com/gin-gonic/gin"
	"go.uber.org/zap"
)

// CompareTimePointers -- Compare the values of two pointers to timestamps, return true if same time or both nil
func CompareTimePointers(one, two *time.Time) (equal bool) {
	if one != nil && two != nil {
		return one.Equal(*two)
	}
	if one == nil && two == nil {
		return true
	}
	return false
}

// FlagEmpty -- Trim and return supplied string, use value of "*UNSET*" if empty string
func FlagEmpty(value string) string {
	check := strings.TrimSpace(value)
	if check == "" {
		check = "*UNSET*"
	}
	return check
}

func ContextGet[T any](ctx *gin.Context, key string) (t T, err error) {
	if temp, found := ctx.Get(key); found {
		var ok bool
		if t, ok = temp.(T); ok {
			return t, nil
		} else {
			return t, fmt.Errorf("failed to convert context value for key %s to %v", key, t)
		}
	}
	return t, fmt.Errorf("failed to find key %s in context", key)
}

// ContextGetUnixMilliFromPtr -- Safely get a timestamp in millis from a context timestamp, or nil if the pointer is nil
func ContextGetUnixMilliFromPtr(ctx *gin.Context, key string) *int64 {
	t, err := ContextGet[*time.Time](ctx, key)
	if err != nil {
		logger.BaseLogger().Debug(fmt.Sprintf("error extracting %s from context", key))
		return nil
	} else if t == nil {
		return nil
	}

	timestamp := t.UnixMilli()
	return &timestamp
}

// EnvOrDefault -- Attempt to get environment variable, or use fallback if not defined
func EnvOrDefault(key, fallback string) string {
	value, found := os.LookupEnv(key)
	if !found {
		value = fallback
	}
	return value
}

func SafeTimeString(value *time.Time) string {
	if value == nil {
		return "nil"
	} else {
		return value.String()
	}
}

const (
	AdditionalLogFields      = "AdditionalLogFields"
	AdditionalDebugLogFields = "AdditionalDebugLogFields"
)

func addAdditionalLogFieldsByKey(c *gin.Context, key string, fields ...zap.Field) {
	if len(fields) == 0 {
		return
	}
	additionalFields, _ := ContextGet[[]zap.Field](c, key)
	c.Set(key, append(additionalFields, fields...))
}

func AddAdditionalDebugLogFields(c *gin.Context, fields ...zap.Field) {
	addAdditionalLogFieldsByKey(c, AdditionalDebugLogFields, fields...)
}

func AddAdditionalLogFields(c *gin.Context, fields ...zap.Field) {
	addAdditionalLogFieldsByKey(c, AdditionalLogFields, fields...)
}

func getAdditionalLogFieldsByKey(c *gin.Context, key string) []zap.Field {
	additionalFields, err := ContextGet[[]zap.Field](c, key)
	if err != nil {
		return []zap.Field{}
	}
	return additionalFields
}

func GetAdditionalDebugLogFields(c *gin.Context) []zap.Field {
	return getAdditionalLogFieldsByKey(c, AdditionalDebugLogFields)
}

func GetAdditionalLogFields(c *gin.Context) []zap.Field {
	return getAdditionalLogFieldsByKey(c, AdditionalLogFields)
}

func ZapTimeOrNull(key string, t *time.Time) zap.Field {
	if t == nil {
		return zap.String(key, "null")
	}
	return zap.Time(key, *t)
}

var deviceContextKey string = "device-context"

func GetDeviceAuthContext(ctx *gin.Context) (*devicecontext.DeviceAuthContext, error) {
	return ContextGet[*devicecontext.DeviceAuthContext](ctx, deviceContextKey)
}

func SetDeviceAuthContext(ctx *gin.Context, authContext *devicecontext.DeviceAuthContext) {
	if authContext == nil {
		return
	}
	ctx.Set(deviceContextKey, authContext)
}

func GetAndValidateDeviceAuthContext(ctx *gin.Context) (*devicecontext.DeviceAuthContext, error) {
	authContext, err := GetDeviceAuthContext(ctx)
	if err != nil {
		return nil, err
	}
	if authContext == nil {
		return nil, errors.New("Auth context was nil")
	}
	return authContext, authContext.Validate()
}

var shouldLogContextKey string = "should-log"

func GetShouldLog(ctx *gin.Context) bool {
	return ctx.GetBool(shouldLogContextKey)
}

func SetShouldLog(ctx *gin.Context, value bool) {
	ctx.Set(shouldLogContextKey, value)
}
