// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package utils

import (
	"testing"
	"time"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

var _ = Describe("Utilities unit tests", func() {
	Context("CompareTimePointers() testing", func() {
		It("should return true if both are nil", func() {
			check := CompareTimePointers(nil, nil)
			Expect(check).To(BeEquivalentTo(true))
		})
		It("should return true if both pointer values have the same time", func() {
			now1 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			now2 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			check := CompareTimePointers(&now1, &now2)
			Expect(check).To(BeEquivalentTo(true))
		})
		It("should return false if both pointer values have different times", func() {
			now1 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			now2 := time.Date(2023, 5, 19, 10, 15, 30, 1, time.UTC) // 1 ns difference
			check := CompareTimePointers(&now1, &now2)
			Expect(check).To(BeEquivalentTo(false))
		})
		It("should return false if one pointer is nil", func() {
			now1 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			check := CompareTimePointers(&now1, nil)
			Expect(check).To(BeEquivalentTo(false))
		})
	})

	Context("FlagEmpty() testing", func() {
		It("should return string unmolested if ok", func() {
			check := FlagEmpty("is_ok")
			Expect(check).To(BeEquivalentTo("is_ok"))
		})
		It("should return string trimmed", func() {
			check := FlagEmpty("   is_trimmed  ")
			Expect(check).To(BeEquivalentTo("is_trimmed"))
		})
		It("should return *UNSET* if all whitespace or empty string", func() {
			check := FlagEmpty("    ")
			Expect(check).To(BeEquivalentTo("*UNSET*"))
			check = FlagEmpty("")
			Expect(check).To(BeEquivalentTo("*UNSET*"))
		})
	})

	Context("EnvOrDefault() testing", func() {
		It("should return value if defined", func() {
			check := EnvOrDefault("PWD", "")
			Expect(check).ToNot(BeEquivalentTo(""))
		})
		It("should return default if not defined", func() {
			check := EnvOrDefault("WONT_BE_DEFINED_I_HOPE", "fallback")
			Expect(check).To(BeEquivalentTo("fallback"))
		})
	})

	// NOTE: Added for coverage
	Context("HeaderStr() testing", func() {
		It("should return padded value", func() {
			check := HeaderStr("Test")
			Expect(check).To(BeEquivalentTo("===== Test ====================================================================="))
		})

		It("large tags always have at least ===== on the other side", func() {
			check := HeaderStr("75--------------------------------------------------------------------chars")
			Expect(check).To(BeEquivalentTo("===== 75--------------------------------------------------------------------chars ====="))
		})
	})
})

func TestUtils(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "All Utils Test Suite")
}
